#!/usr/bin/env python3
# Copyright 2024 Canonical Ltd.
# See LICENSE file for licensing details.

"""This module is only here for backwards compatibility.

Its contents have been moved over to cosl.interfaces.cluster and cosl.interfaces.utils.
"""


from cosl.interfaces.utils import DatabagModel, DataValidationError  # noqa # type:ignore
from cosl.interfaces.cluster import *  # noqa # type:ignore

from logging import getLogger

logger = getLogger("interface-deprecated")
logger.warning(
    "this module has been deprecated and may be removed in a future version: "
    "please use cosl.interfaces.cluster and cosl.interfaces.utils instead."
)



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

"""Helper utils for cosl."""

import importlib
from typing import List


def check_libs_installed(*path: str):
    """Attempt to import these charm libs and raise an error if it fails."""
    libs_not_found: List[str] = []
    for charm_lib_path in path:
        try:
            importlib.import_module(charm_lib_path)
        except ModuleNotFoundError:
            libs_not_found.append(charm_lib_path)

    if libs_not_found:
        install_script = "\n".join(f"charmcraft fetch-lib {libname}" for libname in libs_not_found)
        raise RuntimeError(
            f"Unmet dependencies: the coordinator charm base is missing some charm libs. \
            Please install them with: \n\n{install_script}"
        )



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.
"""Alerting and Recording Rules.

## Overview

## Rules

This library also supports gathering alerting and recording rules from all
related charms and enabling corresponding alerting/recording rules within the
Prometheus charm.  Alert rules are automatically gathered by `AlertRules`
charms when using this library, from a directory conventionally named as one of:
- `prometheus_alert_rules`
- `prometheus_recording_rules`
- `loki_alert_rules`
- `loki_recording_rules`

This directory must reside at the top level in the `src` folder of the consumer
charm. Each file in this directory is assumed to be in one of two formats:
- the official Prometheus rule format, conforming to the
[Prometheus docs](https://prometheus.io/docs/prometheus/latest/configuration/alerting_rules/)
- a single rule format, which is a simplified subset of the official format,
comprising a single alert rule per file, using the same YAML fields.

The file name must have one of the following extensions:
- `.rule`
- `.rules`
- `.yml`
- `.yaml`

An example of the contents of such a file in the custom single rule
format is shown below.

```
alert: HighRequestLatency
expr: job:request_latency_seconds:mean5m{my_key=my_value} > 0.5
for: 10m
labels:
  severity: Medium
  type: HighLatency
annotations:
  summary: High request latency for {{ $labels.instance }}.
```

The `[Alert|Recording]Rules` instance will read all available rules and
also inject "filtering labels" into the expressions. The
filtering labels ensure that rules are localised to the metrics
provider charm's Juju topology (application, model and its UUID). Such
a topology filter is essential to ensure that rules submitted by
one provider charm generates information only for that same charm. When
rules are embedded in a charm, and the charm is deployed as a
Juju application, the rules from that application have their
expressions automatically updated to filter for metrics/logs coming from
the units of that application alone. This removes risk of spurious
evaluation, e.g., when you have multiple deployments of the same charm
monitored by the same Prometheus or Loki.

Not all rules one may want to specify can be embedded in a
charm. Some rules will be specific to a user's use case. This is
the case, for example, of rules that are based on business
constraints, like expecting a certain amount of requests to a specific
API every five minutes. Such alerting or recording rules can be specified
via the [COS Config Charm](https://charmhub.io/cos-configuration-k8s),
which allows importing alert rules and other settings like dashboards
from a Git repository.

Gathering rules and generating rule files within a
charm is easily done using the `alerts()` or `recording_rules()` method(s)
of the consuming charm. Rules generated will automatically include Juju
topology labels. These labels indicate the source of the record or alert.
The following labels are automatically included with each rule:

- `juju_model`
- `juju_model_uuid`
- `juju_application`
"""  # noqa: W505

import hashlib
import logging
import re
from abc import ABC, abstractmethod
from pathlib import Path
from types import SimpleNamespace
from typing import Any, Dict, Final, List, Optional, Union, cast

import yaml

from . import CosTool, JujuTopology
from .types import (
    OfficialRuleFileItem,
    QueryType,
    RuleType,
)

logger = logging.getLogger(__name__)


_generic_alert_rules: Final = SimpleNamespace(
    # We use "5m" to avoid false positives on expected temporary "down", e.g. during intentional (re)start.
    # Juju topology will be later injected by providers of alert rules.
    host_down={
        "alert": "HostDown",
        "expr": "up < 1",
        "for": "5m",
        "labels": {"severity": "critical"},
        "annotations": {
            "summary": "Host '{{ $labels.instance }}' is down.",
            "description": """Host '{{ $labels.instance }}' is down, failed to scrape.
                            VALUE = {{ $value }}
                            LABELS = {{ $labels }}""",
        },
    },
    host_metrics_missing={
        "alert": "HostMetricsMissing",
        # We use "absent(up)" with "for: 5m" because the alert transitions from "Pending" to "Firing".
        # If query portability is desired, "absent_over_time(up[5m])" is an alternative.
        "expr": "absent(up)",
        "for": "5m",
        "labels": {"severity": "critical"},
        "annotations": {
            "summary": "Metrics not received from host '{{ $labels.instance }}', failed to remote write.",
            "description": """Metrics not received from host '{{ $labels.instance }}', failed to remote write.
                            VALUE = {{ $value }}
                            LABELS = {{ $labels }}""",
        },
    },
)

"""
Generic alert rules are in groups to ensure a predictable group name.
"""
generic_alert_groups: Final = SimpleNamespace(
    # Group names must be unique per alert rule file. The final group names may be adjusted by the
    # providers of alert rules to include some topology information, to addresses deduplication.
    application_rules={
        "groups": [
            {
                "name": "HostHealth",
                "rules": [
                    _generic_alert_rules.host_down,
                    _generic_alert_rules.host_metrics_missing,
                ],
            },
        ]
    },
    # If we push to Prometheus via remote-write with an aggregator, there are no UP metrics associated.
    # Only a time series for the metrics we have pushed is available so omit the HostDown rule.
    aggregator_rules={
        "groups": [
            {
                "name": "AggregatorHostHealth",
                "rules": [
                    _generic_alert_rules.host_metrics_missing,
                ],
            },
        ]
    },
)


class InvalidRulePathError(Exception):
    """Raised if the rules folder cannot be found or is otherwise invalid."""

    def __init__(
        self,
        rules_absolute_path: Path,
        message: str,
    ):
        self.rules_absolute_path = rules_absolute_path
        self.message = message

        super().__init__(self.message)


class Rules(ABC):
    """Utility class for amalgamating alerting/recording rule  files and injecting juju topology.

    A `Rules` object supports aggregating rules from files and directories in both
    official and single rule file formats using the `add_path()` method. All the rules
    read are annotated with Juju topology labels and amalgamated into a single data structure
    in the form of a Python dictionary using the `as_dict()` method. Such a dictionary can be
    easily dumped into JSON format and exchanged over relation data. The dictionary can also
    be dumped into YAML format and written directly into a rules file that is read by
    Prometheus. Note that multiple `Rules` objects must not be written into the same file,
    since Prometheus allows only a single list of rule groups per rules file.

    The official  format is a YAML file conforming to the Prometheus/Cortex documentation
    (https://prometheus.io/docs/prometheus/latest/configuration/alerting_rules/).
    The custom single rule format is a subsection of the official YAML, having a single alert
    rule, effectively "one alert per file".
    """

    # This class uses the following terminology for the various parts of a rule file:
    # - rules file: the entire groups[] yaml, including the "groups:" key.
    # - groups (plural): the list of groups[] (a list, i.e. no "groups:" key) - it is a list
    #   of dictionaries that have the "name" and "rules" keys.
    # - group (singular): a single dictionary that has the "name" and "rules" keys.
    # - rules (plural): all the rules in a given group - a list of dictionaries with
    #   the `self.rule_type` (either "alert" or "record") and "expr" keys.
    # - rule (singular): a single dictionary that has the `self.rule_type` (either "alert" or
    #   "record") and "expr" keys.

    def __init__(self, query_type: QueryType, topology: Optional[JujuTopology] = None):
        """Build a rule object.

        Args:
            query_type: either "promql" or "logql" to indicate the query language used
                in the rules, for manipulation with CosTool
            topology: an optional `JujuTopology` instance that is used to annotate all rules.
        """
        self.query_type = query_type
        self.topology = topology
        self.tool = CosTool(default_query_type=query_type)
        self.groups: List[OfficialRuleFileItem] = []

    @property
    @abstractmethod
    def rule_type(self) -> RuleType:
        """Return the rule type being used for interpolation in messages."""
        pass

    # --- HELPER METHODS FOR READING FILES, SHOULD BE STATIC --- #

    @staticmethod
    def _is_official_rule_format(rules_dict: Dict[str, Any]) -> bool:
        """Are rules in the upstream format as supported by Prometheus or Loki.

        Rules in dictionary format are in "official" form if they
        contain a "groups" key, since this implies they contain a list of
        rule groups.

        Args:
            rules_dict: a set of rules in Python dictionary format

        Returns:
            True if rules are in official file format.
        """
        return "groups" in rules_dict

    @staticmethod
    def _is_single_rule_format(rules_dict: Dict[str, Any], rule_type: RuleType) -> bool:
        """Are alert rules in single rule format.

        This library supports reading of rules in a custom format that
        consists of a single rule per file. This does not conform to the
        official rule file format, which requires that each rules file
        consists of a list of rule groups and each group consists of a
        list of rules.

        Rules in dictionary form are considered to be in single rule
        format if in the least it contains two keys corresponding to the
        rule name and expression.

        Returns:
            True if rule is in single rule file format.
        """
        # one rule per file
        return set(rules_dict) >= {rule_type, "expr"}

    @staticmethod
    def _multi_suffix_glob(
        dir_path: Path, suffixes: List[str], recursive: bool = True
    ) -> List[Path]:
        """Helper function for getting all files in a directory that have a matching suffix.

        Args:
            dir_path: path to the directory to glob from.
            suffixes: list of suffixes to include in the glob (items should begin with a period).
            recursive: a flag indicating whether a glob is recursive (nested) or not.

        Returns:
            List of files in `dir_path` that have one of the suffixes specified in `suffixes`.
        """
        all_files_in_dir = dir_path.glob("**/*" if recursive else "*")
        return list(filter(lambda f: f.is_file() and f.suffix in suffixes, all_files_in_dir))

    def _from_dir(self, dir_path: Path, recursive: bool) -> List[OfficialRuleFileItem]:
        """Read all rule files in a directory.

        All rules from files for the same directory are loaded into a single
        group. The generated name of this group includes juju topology.
        By default, only the top directory is scanned; for nested scanning, pass `recursive=True`.

        Args:
            dir_path: directory containing *.rule files (rules without groups).
            recursive: flag indicating whether to scan for rule files recursively.

        Returns:
            a list of dictionaries representing prometheus rule groups, each dictionary
            representing a group (structure determined by `yaml.safe_load`).
        """
        groups: List[OfficialRuleFileItem] = []

        # Gather all records into a list of groups
        for file_path in Rules._multi_suffix_glob(
            dir_path, [".rule", ".rules", ".yml", ".yaml"], recursive
        ):
            groups_from_file = self._from_file(dir_path, file_path)
            if groups_from_file:
                logger.debug("Reading %s rule from %s", self.rule_type, file_path)
                groups.extend(groups_from_file)  # type: ignore

        return groups

    def _from_file(  # noqa: C901
        self, root_path: Path, file_path: Path
    ) -> List[OfficialRuleFileItem]:
        """Read a rules file from path.

        Args:
            root_path: full path to the root rules folder (used only for generating group name)
            file_path: full path to a *.rule file.

        Returns:
            A list of dictionaries representing the rules file, if file is valid (the structure is
            formed by `yaml.safe_load` of the file); an empty list otherwise.
        """
        with file_path.open() as rf:
            # Load a list of rules from file then add labels and filters
            try:
                rule_file = yaml.safe_load(rf)

            except Exception as e:
                logger.error("Failed to read rules from %s: %s", file_path.name, e)
                return []

            # Generate group name prefix
            #  - name, from juju topology
            #  - suffix, from the relative path of the rule file;
            rel_path = file_path.parent.relative_to(root_path)
            rel_path = "" if rel_path == Path(".") else str(rel_path)
            group_name_parts = [self.topology.identifier] if self.topology else []
            group_name_parts.append(rel_path)
            group_name_prefix = "_".join(filter(None, group_name_parts))

            try:
                groups = self._from_dict(
                    rule_file, group_name=file_path.stem, group_name_prefix=group_name_prefix
                )
            except ValueError as e:
                logger.error("Invalid rules file: %s (%s)", file_path.name, e)
                return []

            return groups

    def _from_dict(
        self,
        rule_dict: Dict[str, Any],
        *,
        group_name: Optional[str] = None,
        group_name_prefix: Optional[str] = None,
    ) -> List[OfficialRuleFileItem]:
        """Process rules from dict, injecting juju topology. If a single-rule format is provided, a hash of the yaml file is injected into the group name to ensure uniqueness.

        Args:
            rule_dict: rules content in single-rule or official-rule format as a YAML dict
            group_name: a custom identifier for the rule name to include in the group name
            group_name_prefix: a custom group identifier to prefix the resulting group name, likely Juju topology and relative path context

        Raises:
            ValueError, when invalid rule format given.
        """
        if not rule_dict:
            raise ValueError("Empty")

        if self._is_official_rule_format(rule_dict):
            groups = rule_dict["groups"]
        elif self._is_single_rule_format(rule_dict, self.rule_type):
            if not group_name:
                # Note: the caller of this function should ensure this never happens:
                # Either we use the standard format, or we'd pass a group_name.
                # If/when we drop support for the single-rule-per-file format, this won't
                # be needed anymore.
                group_name = hashlib.shake_256(str(rule_dict).encode("utf-8")).hexdigest(10)

            # convert to list of groups to match official rule format
            groups = [{"name": group_name, "rules": [rule_dict]}]
        else:
            # invalid/unsupported
            raise ValueError("Invalid rule format")

        # update rules with additional metadata
        groups = cast(List[OfficialRuleFileItem], groups)
        for group in groups:
            if not self._is_already_modified(group["name"]):
                # update group name with topology and sub-path
                group["name"] = "_".join(
                    filter(None, [group_name_prefix, group["name"], f"{self.rule_type}s"])
                )
            # after sanitizing we should not modify group["name"] anymore
            group["name"] = self._sanitize_metric_name(group["name"])

            # add "juju_" topology labels
            for rule in group["rules"]:
                if "labels" not in rule:
                    rule["labels"] = {}

                if self.topology:
                    # only insert labels that do not already exist
                    for label, val in self.topology.label_matcher_dict.items():
                        if label not in rule["labels"]:
                            rule["labels"][label] = val

                    # insert juju topology filters into a prometheus rule
                    repl = r'job=~".+"' if self.query_type == "logql" else ""
                    rule["expr"] = self.tool.inject_label_matchers(  # type: ignore
                        expression=re.sub(r"%%juju_topology%%,?", repl, rule["expr"]),
                        topology={
                            k: rule["labels"][k]
                            for k in ("juju_model", "juju_model_uuid", "juju_application")
                            if rule["labels"].get(k) is not None
                        },
                        query_type=self.query_type,
                    )

        return groups

    def _is_already_modified(self, name: str) -> bool:
        """Detect whether a group name has already been modified with juju topology."""
        modified_matcher = re.compile(r"^.*?_[\da-f]{8}_.*?alerts$")
        if modified_matcher.match(name) is None:
            return False
        return True

    def _sanitize_metric_name(self, metric_name: str) -> str:
        """Sanitize a metric name according to https://prometheus.io/docs/concepts/data_model/#metric-names-and-labels."""
        return "".join(char if re.match(r"[a-zA-Z0-9_:]", char) else "_" for char in metric_name)

    # ---- END STATIC HELPER METHODS --- #

    def add(
        self,
        rule_dict: Dict[str, Any],
        group_name: Optional[str] = None,
        group_name_prefix: Optional[str] = None,
    ) -> None:
        """Add rules from dict to the existing ruleset.

        Args:
            rule_dict: a single-rule or official-rule YAML dict
            group_name: a custom group name, used only if the new rule is of single-rule format
            group_name_prefix: a custom group name prefix, used only if the new rule is of single-rule format
        """
        self.groups.extend(
            self._from_dict(rule_dict, group_name=group_name, group_name_prefix=group_name_prefix)
        )

    def add_path(self, dir_path: Union[str, Path], *, recursive: bool = False) -> None:
        """Add rules from a dir path.

        All rules from files are aggregated into a data structure representing a single rule file.
        All group names are augmented with juju topology.

        Args:
            dir_path: either a rules file or a dir of rules files.
            recursive: whether to read files recursively or not (no impact if `path` is a file).
        """
        path = Path(dir_path) if isinstance(dir_path, str) else dir_path
        if path.is_dir():
            self.groups.extend(self._from_dir(path, recursive))
        elif path.is_file():
            self.groups.extend(self._from_file(path.parent, path))  # type: ignore
        else:
            logger.debug("%s rules path does not exist: %s", self.rule_type.capitalize(), path)

    def as_dict(self) -> Dict[str, Any]:
        """Return standard rules file in dict representation.

        Returns:
            a dictionary containing a single list of rule groups.
            The list of rule groups is provided as value of the
            "groups" dictionary key.
        """
        return {"groups": self.groups} if self.groups else {}


class AlertRules(Rules):
    """Utility class for amalgamating alerting files and injecting juju topology.

    The official format is a YAML file conforming to the Prometheus/Cortex documentation
    (https://prometheus.io/docs/prometheus/latest/configuration/alerting_rules/).
    The custom single rule format is a subsection of the official YAML, having a single alert
    rule, effectively "one alert per file".
    """

    _rule_type = "alert"  # type: RuleType

    @property
    def rule_type(self) -> RuleType:
        """Return the rule type being used for interpolation in messages."""
        return self._rule_type


class RecordingRules(Rules):
    """Utility class for amalgamating recording files and injecting juju topology.

    The official format is a YAML file conforming to the Prometheus/Cortex documentation
    (https://prometheus.io/docs/prometheus/latest/configuration/recording_rules/).
    The custom single rule format is a subsection of the official YAML, having a single recording
    rule, effectively "one record per file".
    """

    _rule_type = "record"  # type: RuleType

    @property
    def rule_type(self) -> RuleType:
        """Return the rule type being used for interpolation in messages."""
        return self._rule_type



import dataclasses
import json
from unittest.mock import patch

import ops
import pytest
from ops import RelationChangedEvent, testing

from cosl.coordinated_workers.interface import ClusterRemovedEvent, DataValidationError
from src.cosl.coordinated_workers.coordinator import (
    ClusterRolesConfig,
    Coordinator,
    S3NotFoundError,
)
from src.cosl.interfaces.cluster import ClusterRequirerAppData, ClusterRequirerUnitData
from tests.test_coordinated_workers.test_worker import MyCharm


@pytest.fixture
def coordinator_state():
    requires_relations = {
        endpoint: testing.Relation(endpoint=endpoint, interface=interface["interface"])
        for endpoint, interface in {
            "my-certificates": {"interface": "certificates"},
            "my-logging": {"interface": "loki_push_api"},
            "my-charm-tracing": {"interface": "tracing"},
            "my-workload-tracing": {"interface": "tracing"},
        }.items()
    }
    requires_relations["my-s3"] = testing.Relation(
        "my-s3",
        interface="s3",
        remote_app_data={
            "endpoint": "s3",
            "bucket": "foo-bucket",
            "access-key": "my-access-key",
            "secret-key": "my-secret-key",
        },
    )
    requires_relations["cluster_worker0"] = testing.Relation(
        "my-cluster",
        remote_app_name="worker0",
        remote_app_data=ClusterRequirerAppData(role="read").dump(),
        remote_units_data={
            0: ClusterRequirerUnitData(
                juju_topology={
                    "application": "reader",
                    "unit": "reader/0",
                    "charm_name": "test-reader",
                },
                address="something",
            ).dump()
        },
    )
    requires_relations["cluster_worker1"] = testing.Relation(
        "my-cluster",
        remote_app_name="worker1",
        remote_app_data=ClusterRequirerAppData(role="write").dump(),
        remote_units_data={
            0: ClusterRequirerUnitData(
                juju_topology={
                    "application": "writer",
                    "unit": "writer/0",
                    "charm_name": "test-writer",
                },
                address="something",
            ).dump()
        },
    )
    requires_relations["cluster_worker2"] = testing.Relation(
        "my-cluster",
        remote_app_name="worker2",
        remote_app_data=ClusterRequirerAppData(role="backend").dump(),
        remote_units_data={
            0: ClusterRequirerUnitData(
                juju_topology={
                    "application": "backender",
                    "unit": "backender/0",
                    "charm_name": "test-backender",
                },
                address="something",
            ).dump()
        },
    )

    provides_relations = {
        endpoint: testing.Relation(endpoint=endpoint, interface=interface["interface"])
        for endpoint, interface in {
            "my-dashboards": {"interface": "grafana_dashboard"},
            "my-metrics": {"interface": "prometheus_scrape"},
        }.items()
    }

    return testing.State(
        containers={
            testing.Container("nginx", can_connect=True),
            testing.Container("nginx-prometheus-exporter", can_connect=True),
        },
        relations=list(requires_relations.values()) + list(provides_relations.values()),
    )


@pytest.fixture()
def coordinator_charm(request):
    class MyCoordinator(ops.CharmBase):
        META = {
            "name": "foo-app",
            "requires": {
                "my-certificates": {"interface": "certificates"},
                "my-cluster": {"interface": "cluster"},
                "my-logging": {"interface": "loki_push_api"},
                "my-charm-tracing": {"interface": "tracing", "limit": 1},
                "my-workload-tracing": {"interface": "tracing", "limit": 1},
                "my-s3": {"interface": "s3"},
                "my-ds-exchange-require": {"interface": "grafana_datasource_exchange"},
            },
            "provides": {
                "my-dashboards": {"interface": "grafana_dashboard"},
                "my-metrics": {"interface": "prometheus_scrape"},
                "my-ds-exchange-provide": {"interface": "grafana_datasource_exchange"},
            },
            "containers": {
                "nginx": {"type": "oci-image"},
                "nginx-prometheus-exporter": {"type": "oci-image"},
            },
        }

        _worker_ports = None

        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            # Note: Here it is a good idea not to use context mgr because it is "ops aware"
            self.coordinator = Coordinator(
                charm=self,
                # Roles were take from loki-coordinator-k8s-operator
                roles_config=ClusterRolesConfig(
                    roles={"all", "read", "write", "backend"},
                    meta_roles={"all": {"all", "read", "write", "backend"}},
                    minimal_deployment={
                        "read",
                        "write",
                        "backend",
                    },
                    recommended_deployment={
                        "read": 3,
                        "write": 3,
                        "backend": 3,
                    },
                ),
                external_url="https://foo.example.com",
                worker_metrics_port=123,
                endpoints={
                    "certificates": "my-certificates",
                    "cluster": "my-cluster",
                    "grafana-dashboards": "my-dashboards",
                    "logging": "my-logging",
                    "metrics": "my-metrics",
                    "charm-tracing": "my-charm-tracing",
                    "workload-tracing": "my-workload-tracing",
                    "s3": "my-s3",
                    "send-datasource": "my-ds-exchange-provide",
                    "receive-datasource": "my-ds-exchange-require",
                    "catalogue": None,
                },
                nginx_config=lambda coordinator: f"nginx configuration for {coordinator._charm.meta.name}",
                workers_config=lambda coordinator: f"workers configuration for {coordinator._charm.meta.name}",
                worker_ports=self._worker_ports,
                # nginx_options: Optional[NginxMappingOverrides] = None,
                # is_coherent: Optional[Callable[[ClusterProvider, ClusterRolesConfig], bool]] = None,
                # is_recommended: Optional[Callable[[ClusterProvider, ClusterRolesConfig], bool]] = None,
            )

    return MyCoordinator


def test_worker_roles_subset_of_minimal_deployment(
    coordinator_state: testing.State, coordinator_charm: ops.CharmBase
):
    # Test that the combination of worker roles is a subset of the minimal deployment roles

    # GIVEN a coordinator_charm
    ctx = testing.Context(coordinator_charm, meta=coordinator_charm.META)

    # AND a coordinator_state defining relations to worker charms with incomplete distributed roles
    missing_backend_worker_relation = {
        relation
        for relation in coordinator_state.relations
        if relation.remote_app_name != "worker2"
    }

    # WHEN we process any event
    with ctx(
        ctx.on.update_status(),
        state=dataclasses.replace(coordinator_state, relations=missing_backend_worker_relation),
    ) as mgr:
        charm: coordinator_charm = mgr.charm

        # THEN the deployment is not coherent
        assert not charm.coordinator.is_coherent


def test_worker_ports_published(
    coordinator_state: testing.State,
    coordinator_charm: ops.CharmBase,
):
    ports_per_role = {"read": (10, 22), "write": (1, 2132)}

    def _worker_ports(_, role):
        return ports_per_role.get(role, ())

    coordinator_charm._worker_ports = _worker_ports

    # GIVEN a coordinator_charm that has these ports configured
    ctx = testing.Context(coordinator_charm, meta=coordinator_charm.META)

    # WHEN we process any event
    state_out = ctx.run(
        ctx.on.update_status(),
        state=dataclasses.replace(coordinator_state, leader=True),
    )
    # THEN the ports are correctly distributed to the workers via their relations
    worker_relations = state_out.get_relations("my-cluster")
    for relation in worker_relations:
        remote_worker_role = json.loads(relation.remote_app_data["role"])
        expected_ports = ports_per_role.get(remote_worker_role, ())

        assert set(json.loads(relation.local_app_data.get("worker_ports", ()))) == set(
            expected_ports
        )


def test_without_s3_integration_raises_error(
    coordinator_state: testing.State, coordinator_charm: ops.CharmBase
):
    # Test that a charm without an s3 integration raises S3NotFoundError

    # GIVEN a coordinator charm without an s3 integration
    ctx = testing.Context(coordinator_charm, meta=coordinator_charm.META)
    relations_without_s3 = {
        relation for relation in coordinator_state.relations if relation.endpoint != "my-s3"
    }

    # WHEN we process any event
    with ctx(
        ctx.on.update_status(),
        state=dataclasses.replace(coordinator_state, relations=relations_without_s3),
    ) as mgr:
        # THEN the _s3_config method raises an S3NotFoundError
        with pytest.raises(S3NotFoundError):
            mgr.charm.coordinator._s3_config


@pytest.mark.parametrize("region", (None, "canada"))
@pytest.mark.parametrize("tls_ca_chain", (None, ["my ca chain"]))
@pytest.mark.parametrize("bucket", ("bucky",))
@pytest.mark.parametrize("secret_key", ("foo",))
@pytest.mark.parametrize("access_key", ("foo",))
@pytest.mark.parametrize(
    "endpoint, endpoint_stripped",
    (
        ("example.com", "example.com"),
        ("http://example.com", "example.com"),
        ("https://example.com", "example.com"),
    ),
)
def test_s3_integration(
    coordinator_state: testing.State,
    coordinator_charm: ops.CharmBase,
    region,
    endpoint,
    endpoint_stripped,
    secret_key,
    access_key,
    bucket,
    tls_ca_chain,
):
    # Test that a charm with a s3 integration gives the expected _s3_config

    # GIVEN a coordinator charm with a s3 integration
    ctx = testing.Context(coordinator_charm, meta=coordinator_charm.META)
    s3_relation = coordinator_state.get_relations("my-s3")[0]
    relations_except_s3 = [
        relation for relation in coordinator_state.relations if relation.endpoint != "my-s3"
    ]
    s3_app_data = {
        k: json.dumps(v)
        for k, v in {
            **({"region": region} if region else {}),
            **({"tls-ca-chain": tls_ca_chain} if tls_ca_chain else {}),
            "endpoint": endpoint,
            "access-key": access_key,
            "secret-key": secret_key,
            "bucket": bucket,
        }.items()
    }

    # WHEN we process any event
    with ctx(
        ctx.on.update_status(),
        state=dataclasses.replace(
            coordinator_state,
            relations=relations_except_s3
            + [dataclasses.replace(s3_relation, remote_app_data=s3_app_data)],
        ),
    ) as mgr:
        # THEN the s3_connection_info method returns the expected data structure
        coordinator: Coordinator = mgr.charm.coordinator
        assert coordinator.s3_connection_info.region == region
        assert coordinator.s3_connection_info.bucket == bucket
        assert coordinator.s3_connection_info.endpoint == endpoint
        assert coordinator.s3_connection_info.secret_key == secret_key
        assert coordinator.s3_connection_info.access_key == access_key
        assert coordinator.s3_connection_info.tls_ca_chain == tls_ca_chain
        assert coordinator._s3_config["endpoint"] == endpoint_stripped
        assert coordinator._s3_config["insecure"] is (not tls_ca_chain)


def test_tracing_receivers_urls(
    coordinator_state: testing.State, coordinator_charm: ops.CharmBase
):
    charm_tracing_relation = testing.Relation(
        endpoint="my-charm-tracing",
        remote_app_data={
            "receivers": json.dumps(
                [{"protocol": {"name": "otlp_http", "type": "http"}, "url": "1.2.3.4:4318"}]
            )
        },
    )
    workload_tracing_relation = testing.Relation(
        endpoint="my-workload-tracing",
        remote_app_data={
            "receivers": json.dumps(
                [
                    {"protocol": {"name": "otlp_http", "type": "http"}, "url": "5.6.7.8:4318"},
                    {"protocol": {"name": "otlp_grpc", "type": "grpc"}, "url": "5.6.7.8:4317"},
                ]
            )
        },
    )
    ctx = testing.Context(coordinator_charm, meta=coordinator_charm.META)
    with ctx(
        ctx.on.update_status(),
        state=dataclasses.replace(
            coordinator_state, relations=[charm_tracing_relation, workload_tracing_relation]
        ),
    ) as mgr:
        coordinator: Coordinator = mgr.charm.coordinator
        assert coordinator._charm_tracing_receivers_urls == {
            "otlp_http": "1.2.3.4:4318",
        }
        assert coordinator._workload_tracing_receivers_urls == {
            "otlp_http": "5.6.7.8:4318",
            "otlp_grpc": "5.6.7.8:4317",
        }


@pytest.mark.parametrize(
    "event",
    (
        testing.CharmEvents.update_status(),
        testing.CharmEvents.start(),
        testing.CharmEvents.install(),
        testing.CharmEvents.config_changed(),
    ),
)
def test_invalid_databag_content(coordinator_charm: ops.CharmBase, event):
    # Test Invalid relations databag for ClusterProvider.gather_addresses_by_role

    # GIVEN a coordinator charm with a cluster relation and invalid remote databag contents
    requires_relations = {
        endpoint: testing.Relation(endpoint=endpoint, interface=interface["interface"])
        for endpoint, interface in {
            "my-certificates": {"interface": "certificates"},
            "my-logging": {"interface": "loki_push_api"},
            "my-charm-tracing": {"interface": "tracing"},
            "my-workload-tracing": {"interface": "tracing"},
        }.items()
    }
    requires_relations["cluster_worker0"] = testing.Relation(
        "my-cluster",
        remote_app_name="worker0",
        remote_app_data=ClusterRequirerAppData(role="read").dump(),
    )
    requires_relations["cluster_worker1"] = testing.Relation(
        "my-cluster",
        remote_app_name="worker1",
        remote_app_data=ClusterRequirerAppData(role="read").dump(),
    )
    requires_relations["cluster_worker2"] = testing.Relation(
        "my-cluster",
        remote_app_name="worker2",
    )

    provides_relations = {
        endpoint: testing.Relation(endpoint=endpoint, interface=interface["interface"])
        for endpoint, interface in {
            "my-dashboards": {"interface": "grafana_dashboard"},
            "my-metrics": {"interface": "prometheus_scrape"},
        }.items()
    }

    invalid_databag_state = testing.State(
        containers={
            testing.Container("nginx", can_connect=True),
            testing.Container("nginx-prometheus-exporter", can_connect=True),
        },
        relations=list(requires_relations.values()) + list(provides_relations.values()),
    )

    # WHEN: the coordinator processes any event
    ctx = testing.Context(coordinator_charm, meta=coordinator_charm.META)
    with ctx(event, invalid_databag_state) as manager:
        cluster = manager.charm.coordinator.cluster
        # THEN the coordinator sets unit to blocked since the cluster is inconsistent with the missing relation.
        cluster.gather_addresses_by_role()
        manager.run()
    assert cluster.model.unit.status == ops.BlockedStatus("[consistency] Cluster inconsistent.")


@pytest.mark.parametrize("app", (True, False))
def test_invalid_app_or_unit_databag(
    coordinator_charm: ops.CharmBase, coordinator_state, app: bool
):
    # Test that when a relation changes and either the app or unit data is invalid
    #   the worker emits a ClusterRemovedEvent

    # WHEN you define a properly configured charm
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )

    # IF the relation data is invalid (forced by the patched Exception)
    object_to_patch = (
        "cosl.coordinated_workers.interface.ClusterProviderAppData.load"
        if app
        else "cosl.coordinated_workers.interface.ClusterRequirerUnitData.load"
    )

    with patch(object_to_patch, side_effect=DataValidationError("Mock error")):
        # AND the relation changes
        relation = testing.Relation("cluster")

        ctx.run(
            ctx.on.relation_changed(relation),
            testing.State(
                containers={testing.Container("foo", can_connect=True)}, relations={relation}
            ),
        )

    # NOTE: this difference should not exist, and the ClusterRemovedEvent should always
    #   be emitted in case of corrupted data

    # THEN the charm emits a ClusterRemovedEvent
    if app:
        assert len(ctx.emitted_events) == 2
        assert isinstance(ctx.emitted_events[0], RelationChangedEvent)
        assert isinstance(ctx.emitted_events[1], ClusterRemovedEvent)
    else:
        assert len(ctx.emitted_events) == 1
        assert isinstance(ctx.emitted_events[0], RelationChangedEvent)



# Copyright 2020 Canonical Ltd.
# See LICENSE file for licensing details.

import json
import re
import unittest
import uuid
from pathlib import Path

import yaml
from deepdiff import DeepDiff
from fs.tempfs import TempFS

from cosl.juju_topology import JujuTopology
from cosl.rules import AlertRules


class TestEndpointProvider(unittest.TestCase):
    def test_each_alert_rule_is_topology_labeled(self):
        ri = AlertRules(
            query_type="promql",
            topology=JujuTopology(
                model="unittest",
                model_uuid=str(uuid.uuid4()),
                unit="tester/0",
                application="tester",
            ),
        )
        ri.add_path(Path(__file__).resolve().parent / "promql_rules" / "prometheus_alert_rules")

        alerts = ri.as_dict()
        self.assertIn("groups", alerts)
        self.assertEqual(len(alerts["groups"]), 5)
        for group in alerts["groups"]:
            for rule in group["rules"]:
                if "and_unit" not in group["name"]:
                    self.assertIn("labels", rule)
                    labels = rule["labels"]
                    self.assertIn("juju_model", labels)
                    self.assertIn("juju_application", labels)
                    self.assertIn("juju_model_uuid", labels)
                    # alerts should not have unit information if not already present
                    # https://discourse.charmhub.io/t/juju-topology-labels/8874
                    self.assertNotIn("juju_unit", rule["labels"])
                    self.assertNotIn("juju_unit=", rule["expr"])
                else:
                    self.assertIn("labels", rule)
                    labels = rule["labels"]
                    self.assertIn("juju_model", labels)
                    self.assertIn("juju_application", labels)
                    self.assertIn("juju_model_uuid", labels)
                    # unit information is already present
                    self.assertIn("juju_unit", rule["labels"])
                    self.assertIn("juju_unit=", rule["expr"])

    def test_each_alert_expression_is_topology_labeled(self):
        ri = AlertRules(
            query_type="promql",
            topology=JujuTopology(
                model="unittest",
                model_uuid=str(uuid.uuid4()),
                unit="tester/0",
                application="tester",
            ),
        )
        ri.add_path(Path(__file__).resolve().parent / "promql_rules" / "prometheus_alert_rules")

        alerts = ri.as_dict()
        self.assertIn("groups", alerts)
        self.assertEqual(len(alerts["groups"]), 5)
        group = alerts["groups"][0]
        for rule in group["rules"]:
            self.assertIn("expr", rule)
            for labels in expression_labels(rule["expr"]):
                self.assertIn("juju_model", labels)
                self.assertIn("juju_model_uuid", labels)
                self.assertIn("juju_application", labels)


class TestAddRuleFromStr(unittest.TestCase):
    def setUp(self):
        self.rule = {
            "alert": "SomeRuleName",
            "expr": "up < 1",
            "labels": {"severity": "critical"},
        }

    def test_official_rule_add_alerts_from_dict(self):
        official_rule = {
            "groups": [
                {
                    "name": "SomeGroupName",
                    "rules": [
                        {
                            "alert": "FooRule",
                            "expr": "up < 1",
                            "labels": {"severity": "critical"},
                        },
                        {
                            "alert": "BarRule",
                            "expr": "absent(up)",
                            "labels": {"severity": "critical"},
                        },
                    ],
                }
            ]
        }
        expected_rules = {
            "groups": [
                {
                    "name": "initial_alerts",
                    "rules": [self.rule],
                },
                {"name": "SomeGroupName_alerts", "rules": official_rule["groups"][0]["rules"]},
            ]
        }
        # GIVEN an alert rule
        rules = AlertRules(query_type="promql")
        rules.groups = rules._from_dict(self.rule, group_name="initial")
        # WHEN a rule is added from string in official-rule format
        rules.add(official_rule)
        rules_dict = rules.as_dict()
        # THEN the new rule is a combination of all
        self.assertEqual({}, DeepDiff(expected_rules, rules_dict))

    def test_single_rule_add_alerts_from_dict(self):
        single_rule = {
            "alert": "BarRule",
            "expr": "up < 1",
            "labels": {"severity": "critical"},
        }
        expected_rules = {
            "groups": [
                {
                    "name": "initial_alerts",
                    "rules": [self.rule],
                },
                {"name": "some_new_alerts", "rules": [single_rule]},
            ]
        }
        # GIVEN an alert rule
        rules = AlertRules(query_type="promql")
        rules.groups = rules._from_dict(self.rule, group_name="initial")
        # WHEN a rule is added from string in single-rule format with a custom group name and prefix
        rules.add(
            single_rule,
            group_name="new",
            group_name_prefix="some",
        )
        # THEN the new rule is a combination of all
        rules_dict = rules.as_dict()
        # AND the added rule group name has the custom group name and prefix
        self.assertEqual({}, DeepDiff(expected_rules, rules_dict))


class TestFromDictGroupName(unittest.TestCase):
    def setUp(self):
        self.official_rule = {
            "groups": [
                {
                    "name": "SomeGroupName",
                    "rules": [
                        {
                            "alert": "SomeRuleName",
                            "expr": "up < 1",
                            "labels": {"severity": "critical"},
                        }
                    ],
                }
            ]
        }
        self.single_rule = {
            "alert": "SomeRuleName",
            "expr": "up < 1",
            "labels": {"severity": "critical"},
        }

    def test_single_rule_from_dict_group_sanitized(self):
        # GIVEN an alert rule in single-rule format
        rules = AlertRules(query_type="promql")
        # WHEN processed as string and provided an illegal custom group name
        groups = rules._from_dict(
            self.single_rule, group_name="Foo$123/Hello:World(go_od)bye!@#^&*()[]{}|;:,.<>?`~_"
        )
        for group in groups:
            # THEN the group name only contains characters in [a-zA-Z0-9_:]
            # AND special characters are replaced with "_"
            self.assertEqual(
                group["name"], "Foo_123_Hello:World_go_od_bye______________:_________alerts"
            )

    def test_single_rule_from_dict(self):
        # GIVEN an alert rule in single-rule format
        rules = AlertRules(query_type="promql")
        # WHEN processed as string
        groups = rules._from_dict(self.single_rule)
        for group in groups:
            # THEN group name contains hash
            self.assertNotEqual(get_hash(group["name"]), "")

    def test_single_rule_from_dict_custom_group(self):
        # GIVEN an alert rule in single-rule format
        rules = AlertRules(query_type="promql")
        # WHEN processed as string and provided custom group name
        groups = rules._from_dict(self.single_rule, group_name="foo")
        for group in groups:
            # THEN group name does not contain hash
            self.assertEqual(get_hash(group["name"]), "")
            # AND group name contains the custom group name
            self.assertIn("foo", group["name"])

    def test_single_rule_from_dict_custom_group_prefix(self):
        # GIVEN an alert rule in single-rule format
        rules = AlertRules(query_type="promql")
        # WHEN processed as string and provided custom group name prefix
        groups = rules._from_dict(self.single_rule, group_name_prefix="foo")
        for group in groups:
            # THEN group name does not include the original group name
            self.assertNotIn("SomeGroupName", group["name"])
            # AND group name is prefixed with custom value
            self.assertTrue(group["name"].startswith("foo"))

    def test_official_rule_from_dict(self):
        # GIVEN an alert rule in official-rule format
        rules = AlertRules(query_type="promql")
        # WHEN processed as dict
        groups = rules._from_dict(self.official_rule)
        for group in groups:
            # THEN group name matches the group name in the alert
            self.assertIn("SomeGroupName", group["name"])

    def test_official_rule_from_dict_custom_group_prefix(self):
        # GIVEN an alert rule in official-rule format
        rules = AlertRules(query_type="promql")
        # WHEN processed as string and provided custom group name prefix
        groups = rules._from_dict(self.official_rule, group_name_prefix="foo")
        for group in groups:
            # THEN group name includes the original group name
            self.assertIn("SomeGroupName", group["name"])
            # AND group name is prefixed with custom value
            self.assertTrue(group["name"].startswith("foo"))

    def test_raises_value_error_empty_dict(self):
        rules = AlertRules(query_type="promql")
        with self.assertRaises(ValueError) as ctx:
            rules._from_dict("")
        self.assertEqual(str(ctx.exception), "Empty")

    def test_raises_value_error_invalid_rule_format(self):
        invalid_rule = {
            "expr": "up < 1",
            "labels": {"severity": "critical"},
        }
        rules = AlertRules(query_type="promql")
        with self.assertRaises(ValueError) as ctx:
            rules._from_dict(invalid_rule)
        self.assertEqual(str(ctx.exception), "Invalid rule format")


def sorted_matchers(matchers) -> str:
    parts = [m.strip() for m in matchers.split(",")]
    return ",".join(sorted(parts))


def expression_labels(expr):
    """Extract labels from an alert rule expression.

    Args:
        expr: a string representing an alert expression.

    Returns:
        a generator which yields each set of labels in
        in the expression.
    """
    pattern = re.compile(r"\{.*\}")
    matches = pattern.findall(expr)
    for match in matches:
        match = match.replace("=", '":').replace("juju_", '"juju_')
        labels = json.loads(match)
        yield labels


def get_hash(group_name: str) -> str:
    """Extract the hash of the group name when a group name was not provided to _from_dict method in the Rules class.

    This occurs when the single-rule format is provided rather than official-rule format.

    Args:
        group_name: a string representing the group name of the rules.

    Returns:
        the matching hexdigest of the group name if a match is found, otherwise an empty string.
    """
    import re

    pattern = r"([a-f0-9]{20})_"
    match = re.search(pattern, group_name)
    if match:
        return match.group(1)
    else:
        return ""


class TestAlertRulesWithOneRulePerFile(unittest.TestCase):
    def setUp(self) -> None:
        free_standing_rule = {
            "alert": "free_standing",
            "expr": "avg(some_vector[5m]) > 5",
        }

        alert_rule = {
            "alert": "CPUOverUse",
            "expr": "process_cpu_seconds_total{%%juju_topology%%} > 0.12",
        }
        rules_file_dict = {"groups": [{"name": "group1", "rules": [alert_rule]}]}

        self.sandbox = TempFS("rule_files", auto_clean=True)
        self.addCleanup(self.sandbox.close)
        self.sandbox.makedirs("rules/prom/mixed_format")
        self.sandbox.makedirs("rules/prom/lma_format")
        self.sandbox.makedirs("rules/prom/prom_format")
        self.sandbox.writetext("rules/prom/mixed_format/lma_rule.rule", yaml.safe_dump(alert_rule))
        self.sandbox.writetext(
            "rules/prom/mixed_format/standard_rule.rule", yaml.safe_dump(rules_file_dict)
        )
        self.sandbox.writetext(
            "rules/prom/lma_format/free_standing_rule.rule", yaml.safe_dump(free_standing_rule)
        )
        self.sandbox.writetext(
            "rules/prom/prom_format/standard_rule.rule", yaml.safe_dump(rules_file_dict)
        )

        self.topology = JujuTopology(
            "MyModel", "12de4fae-06cc-4ceb-9089-567be09fec78", "MyApp", "MyUnit", "MyCharm"
        )

    def test_non_recursive_is_default(self):
        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(self.sandbox.getsyspath("/rules/prom/"))
        rules_file_dict = rules.as_dict()
        self.assertEqual({}, rules_file_dict)

    def test_non_recursive_lma_format_loading_from_root_dir(self):
        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(self.sandbox.getsyspath("/rules/prom/lma_format/"))
        rules_file_dict = rules.as_dict()

        expected_freestanding_rule = {
            "alert": "free_standing",
            "expr": "avg(some_vector[5m]) > 5",
            "labels": self.topology.label_matcher_dict,
        }

        expected_rules_file = {
            "groups": [
                {
                    "name": f"{sorted_matchers(self.topology.identifier)}_free_standing_rule_alerts",
                    "rules": [expected_freestanding_rule],
                },
            ]
        }

        self.assertEqual(expected_rules_file, rules_file_dict)

    def test_non_recursive_official_format_loading_from_root_dir(self):
        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(self.sandbox.getsyspath("/rules/prom/prom_format"))
        rules_file_dict = rules.as_dict()

        expected_alert_rule = {
            "alert": "CPUOverUse",
            "expr": f"process_cpu_seconds_total{{{sorted_matchers(self.topology.alert_expression_str)}}} > 0.12",
            "labels": self.topology.label_matcher_dict,
        }

        expected_rules_file = {
            "groups": [
                {
                    "name": f"{self.topology.identifier}_group1_alerts",
                    "rules": [expected_alert_rule],
                },
            ]
        }
        self.assertEqual(expected_rules_file, rules_file_dict)

    def test_alerts_in_both_formats_are_recursively_aggregated(self):
        """This test covers several aspects of the rules format.

        - Group name:
          - For rules in lma format, core group name is the filename
          - For rules in official format, core group name is the group name in the file
        """
        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(self.sandbox.getsyspath("/rules/prom"), recursive=True)
        rules_file_dict = rules.as_dict()

        expected_alert_rule = {
            "alert": "CPUOverUse",
            "expr": f"process_cpu_seconds_total{{{sorted_matchers(self.topology.alert_expression_str)}}} > 0.12",
            "labels": self.topology.label_matcher_dict,
        }

        expected_freestanding_rule = {
            "alert": "free_standing",
            "expr": "avg(some_vector[5m]) > 5",
            "labels": self.topology.label_matcher_dict,
        }

        expected_rules_file = {
            "groups": [
                {
                    "name": f"{self.topology.identifier}_mixed_format_group1_alerts",
                    "rules": [expected_alert_rule],
                },
                {
                    "name": f"{self.topology.identifier}_mixed_format_lma_rule_alerts",
                    "rules": [expected_alert_rule],
                },
                {
                    "name": f"{self.topology.identifier}_lma_format_free_standing_rule_alerts",
                    "rules": [expected_freestanding_rule],
                },
                {
                    "name": f"{self.topology.identifier}_prom_format_group1_alerts",
                    "rules": [expected_alert_rule],
                },
            ]
        }

        self.assertEqual({}, DeepDiff(expected_rules_file, rules_file_dict, ignore_order=True))

    def test_unit_not_in_alert_labels(self):
        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(self.sandbox.getsyspath("/rules/prom"), recursive=True)
        rules_file_dict = rules.as_dict()
        for group in rules_file_dict["groups"]:
            for rule in group["rules"]:
                self.assertTrue("juju_unit" not in rule["labels"])

    def test_charm_not_in_alert_expression(self):
        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(self.sandbox.getsyspath("/rules/prom"), recursive=True)
        rules_file_dict = rules.as_dict()
        for group in rules_file_dict["groups"]:
            for rule in group["rules"]:
                self.assertTrue("charm=" not in rule["expr"])


class TestAlertRulesWithMultipleRulesPerFile(unittest.TestCase):
    def setUp(self) -> None:
        self.topology = JujuTopology(
            "MyModel", "12de4fae-06cc-4ceb-9089-567be09fec78", "MyApp", "MyCharm"
        )

    def gen_rule(self, name, **extra):
        return {
            "alert": f"CPUOverUse_{name}",
            "expr": f"process_cpu_seconds_total{{{sorted_matchers(self.topology.label_matchers)}}} > 0.12",
            **extra,
        }

    def gen_group(self, name):
        return {"name": f"group_{name}", "rules": [self.gen_rule(1), self.gen_rule(2)]}

    def test_load_multiple_rules_per_file(self):
        """Test official format with multiple alert rules per group in multiple groups."""
        rules_file_dict = {"groups": [self.gen_group(1), self.gen_group(2)]}
        sandbox = TempFS("rule_files", auto_clean=True)
        sandbox.makedirs("rules")
        sandbox.writetext("rules/file.rule", yaml.safe_dump(rules_file_dict))

        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(sandbox.getsyspath("/rules"), recursive=False)
        rules_file_dict_read = rules.as_dict()

        expected_rules_file = {
            "groups": [
                {
                    "name": f"{self.topology.identifier}_group_1_alerts",
                    "rules": [
                        self.gen_rule(1, labels=self.topology.label_matcher_dict),
                        self.gen_rule(2, labels=self.topology.label_matcher_dict),
                    ],
                },
                {
                    "name": f"{self.topology.identifier}_group_2_alerts",
                    "rules": [
                        self.gen_rule(1, labels=self.topology.label_matcher_dict),
                        self.gen_rule(2, labels=self.topology.label_matcher_dict),
                    ],
                },
            ]
        }
        self.assertDictEqual(expected_rules_file, rules_file_dict_read)

    def test_duplicated_alert_names_within_alert_rules_list_are_silently_accepted(self):
        """Test official format when the alert rules list has a duplicated alert name."""
        rules_file_dict = {
            "groups": [
                {
                    "name": "my_group",
                    "rules": [self.gen_rule("same"), self.gen_rule("same")],
                }
            ]
        }
        sandbox = TempFS("rule_files", auto_clean=True)
        sandbox.makedirs("rules")
        sandbox.writetext("rules/file.rule", yaml.safe_dump(rules_file_dict))

        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(sandbox.getsyspath("/rules"), recursive=False)
        rules_file_dict_read = rules.as_dict()

        expected_rules_file = {
            "groups": [
                {
                    "name": f"{self.topology.identifier}_my_group_alerts",
                    "rules": [
                        self.gen_rule("same", labels=self.topology.label_matcher_dict),
                        self.gen_rule("same", labels=self.topology.label_matcher_dict),
                    ],
                },
            ]
        }
        self.assertDictEqual(expected_rules_file, rules_file_dict_read)

    def test_duplicated_group_names_within_a_file_are_silently_accepted(self):
        rules_file_dict = {"groups": [self.gen_group("same"), self.gen_group("same")]}
        sandbox = TempFS("rule_files", auto_clean=True)
        sandbox.makedirs("rules")
        sandbox.writetext("rules/file.rule", yaml.safe_dump(rules_file_dict))

        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(sandbox.getsyspath("/rules"), recursive=False)
        rules_file_dict_read = rules.as_dict()

        expected_rules_file = {
            "groups": [
                {
                    "name": f"{self.topology.identifier}_group_same_alerts",
                    "rules": [
                        self.gen_rule(1, labels=self.topology.label_matcher_dict),
                        self.gen_rule(2, labels=self.topology.label_matcher_dict),
                    ],
                },
                {
                    "name": f"{self.topology.identifier}_group_same_alerts",
                    "rules": [
                        self.gen_rule(1, labels=self.topology.label_matcher_dict),
                        self.gen_rule(2, labels=self.topology.label_matcher_dict),
                    ],
                },
            ]
        }
        self.assertDictEqual(expected_rules_file, rules_file_dict_read)

    def test_deeply_nested(self):
        sandbox = TempFS("rule_files", auto_clean=True)
        sandbox.makedirs("rules/a/b/")
        sandbox.writetext("rules/file.rule", yaml.safe_dump(self.gen_rule(0)))
        sandbox.writetext("rules/a/file.rule", yaml.safe_dump(self.gen_rule(1)))
        sandbox.writetext("rules/a/b/file.rule", yaml.safe_dump(self.gen_rule(2)))

        rules = AlertRules(query_type="promql", topology=self.topology)
        rules.add_path(sandbox.getsyspath("/rules"), recursive=True)
        rules_file_dict_read = rules.as_dict()

        expected_rules_file = {
            "groups": [
                {
                    "name": f"{self.topology.identifier}_file_alerts",
                    "rules": [self.gen_rule(0, labels=self.topology.label_matcher_dict)],
                },
                {
                    "name": f"{self.topology.identifier}_a_file_alerts",
                    "rules": [self.gen_rule(1, labels=self.topology.label_matcher_dict)],
                },
                {
                    "name": f"{self.topology.identifier}_a_b_file_alerts",
                    "rules": [self.gen_rule(2, labels=self.topology.label_matcher_dict)],
                },
            ]
        }
        self.assertDictEqual(expected_rules_file, rules_file_dict_read)


class TestAlertRulesContainingUnitTopology(unittest.TestCase):
    """Tests that check we does not remove unit topology.

    Unit Topology information is not added to alert rules expressions and labels,
    by the Provider. However, if unit topology information is
    present in the labels then it must not be removed since the client that
    the alert be limited to a specific unit.
    """

    def test_unit_label_is_retained_if_hard_coded(self):
        ri = AlertRules(
            query_type="promql",
            topology=JujuTopology(
                model="unittest",
                model_uuid=str(uuid.uuid4()),
                unit="tester/0",
                application="tester",
            ),
        )
        ri.add_path(
            Path(__file__).resolve().parent / "promql_rules" / "alert_rules_with_unit_topology"
        )

        alert_rules = ri.as_dict()
        for group in alert_rules["groups"]:
            for rule in group["rules"]:
                self.assertIn("juju_unit", rule["labels"])
                self.assertIn("juju_unit=", rule["expr"])



import json
from contextlib import ExitStack
from pathlib import Path
from unittest.mock import MagicMock, patch

import ops
import pytest
import yaml
from ops import BlockedStatus, testing
from scenario.context import CharmEvents
from scenario.errors import UncaughtCharmError

from cosl.coordinated_workers.worker import (
    CERT_FILE,
    CLIENT_CA_FILE,
    CONFIG_FILE,
    KEY_FILE,
    S3_TLS_CA_CHAIN_FILE,
    Worker,
)
from tests.test_coordinated_workers.test_worker_status import k8s_patch


class MyCharm(ops.CharmBase):
    layer = ops.pebble.Layer("")

    def __init__(self, framework: ops.Framework):
        super().__init__(framework)
        self.worker = Worker(
            self,
            "foo",
            lambda _: self.layer,
            {"cluster": "cluster"},
            readiness_check_endpoint="http://localhost:3200/ready",
        )


def test_no_roles_error():
    # Test that a charm that defines NO 'role-x' config options, when run,
    # raises a WorkerError

    # WHEN you define a charm with no role-x config options
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={},
    )

    # IF the charm executes any event
    # THEN the charm raises an error
    with pytest.raises(testing.errors.UncaughtCharmError):
        ctx.run(ctx.on.update_status(), testing.State(containers={testing.Container("foo")}))


@pytest.mark.parametrize(
    "roles_active, roles_inactive, expected",
    (
        (
            ["read", "write", "ingester", "all"],
            ["alertmanager"],
            ["read", "write", "ingester", "all"],
        ),
        (["read", "write"], ["alertmanager"], ["read", "write"]),
        (["read"], ["alertmanager", "write", "ingester", "all"], ["read"]),
        ([], ["read", "write", "ingester", "all", "alertmanager"], []),
    ),
)
def test_roles_from_config(roles_active, roles_inactive, expected):
    # Test that a charm that defines any 'role-x' config options, when run,
    # correctly determines which ones are enabled through the Worker

    # WHEN you define a charm with a few role-x config options
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={
            "options": {
                f"role-{r}": {"type": "boolean", "default": "false"}
                for r in (roles_active + roles_inactive)
            }
        },
    )

    # AND the charm runs with a few of those set to true, the rest to false
    with ctx(
        ctx.on.update_status(),
        testing.State(
            containers={testing.Container("foo")},
            config={
                **{f"role-{r}": False for r in roles_inactive},
                **{f"role-{r}": True for r in roles_active},
            },
        ),
    ) as mgr:
        # THEN the Worker.roles method correctly returns the list of only those that are set to true
        assert set(mgr.charm.worker.roles) == set(expected)


@patch.object(Worker, "is_ready", new=lambda _: True)
def test_worker_restarts_if_some_service_not_up(tmp_path):
    # GIVEN a worker with some services
    MyCharm.layer = ops.pebble.Layer(
        {
            "services": {
                "foo": {
                    "summary": "foos all the things",
                    "description": "bar",
                    "startup": "enabled",
                    "override": "merge",
                    "command": "ls -la",
                },
                "bar": {
                    "summary": "bars the foos",
                    "description": "bar",
                    "startup": "enabled",
                    "command": "exit 1",
                },
                "baz": {
                    "summary": "bazzes all of the bars",
                    "description": "bar",
                    "startup": "enabled",
                    "command": "echo hi",
                },
            }
        }
    )
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    # WHEN the charm receives any event and there are no changes to the config or the layer,
    #  but some of the services are down
    cfg = tmp_path / "cfg.yaml"
    cfg.write_text("some: yaml")
    container = testing.Container(
        "foo",
        can_connect=True,
        mounts={"local": testing.Mount(location=CONFIG_FILE, source=cfg)},
        execs={
            testing.Exec(("update-ca-certificates", "--fresh")),
            testing.Exec(("/bin/foo", "-version"), stdout="foo"),
        },
        service_statuses={
            "foo": ops.pebble.ServiceStatus.INACTIVE,
            "bar": ops.pebble.ServiceStatus.ACTIVE,
            "baz": ops.pebble.ServiceStatus.INACTIVE,
        },
    )
    state_out = ctx.run(ctx.on.pebble_ready(container), testing.State(containers={container}))

    # THEN the charm restarts all the services that are down
    container_out = state_out.get_container("foo")
    service_statuses = container_out.service_statuses.values()
    assert all(svc is ops.pebble.ServiceStatus.ACTIVE for svc in service_statuses), [
        stat.value for stat in service_statuses
    ]


@patch.object(Worker, "is_ready", new=lambda _: True)
def test_worker_does_not_restart_external_services(tmp_path):
    # GIVEN a worker with some services and a layer with some other services
    MyCharm.layer = ops.pebble.Layer(
        {
            "services": {
                "foo": {
                    "summary": "foos all the things",
                    "override": "merge",
                    "description": "bar",
                    "startup": "enabled",
                    "command": "ls -la",
                }
            }
        }
    )
    other_layer = ops.pebble.Layer(
        {
            "services": {
                "bar": {
                    "summary": "bars the foos",
                    "description": "bar",
                    "startup": "enabled",
                    "command": "exit 1",
                },
                "baz": {
                    "summary": "bazzes all of the bars",
                    "description": "bar",
                    "startup": "enabled",
                    "command": "echo hi",
                },
            }
        }
    )

    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    # WHEN the charm receives any event and there are no changes to the config or the layer,
    #  but some of the services are down
    cfg = tmp_path / "cfg.yaml"
    cfg.write_text("some: yaml")
    container = testing.Container(
        "foo",
        execs={
            testing.Exec(("update-ca-certificates", "--fresh")),
            testing.Exec(("/bin/foo", "-version"), stdout="foo"),
        },
        can_connect=True,
        mounts={"local": testing.Mount(location=CONFIG_FILE, source=cfg)},
        layers={"foo": MyCharm.layer, "bar": other_layer},
        service_statuses={
            # layer foo has some inactive
            "foo": ops.pebble.ServiceStatus.INACTIVE,
            # layer bar has some inactive
            "bar": ops.pebble.ServiceStatus.ACTIVE,
            "baz": ops.pebble.ServiceStatus.INACTIVE,
        },
    )
    state_out = ctx.run(ctx.on.pebble_ready(container), testing.State(containers={container}))

    # THEN the charm restarts all the services that are down
    container_out = state_out.get_container("foo")
    assert container_out.service_statuses == {
        # layer foo service is now active
        "foo": ops.pebble.ServiceStatus.ACTIVE,
        # layer bar services is unchanged
        "bar": ops.pebble.ServiceStatus.ACTIVE,
        "baz": ops.pebble.ServiceStatus.INACTIVE,
    }


def test_worker_raises_if_service_restart_fails_for_too_long(tmp_path):
    # GIVEN a worker with some services
    MyCharm.layer = ops.pebble.Layer(
        {
            "services": {
                "foo": {
                    "summary": "foos all the things",
                    "description": "bar",
                    "startup": "enabled",
                    "command": "ls -la",
                },
            }
        }
    )
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    cfg = tmp_path / "cfg.yaml"
    cfg.write_text("some: yaml")
    container = testing.Container(
        "foo",
        can_connect=True,
        mounts={"local": testing.Mount(location=CONFIG_FILE, source=cfg)},
        service_statuses={
            "foo": ops.pebble.ServiceStatus.INACTIVE,
        },
    )

    def raise_change_error(*args):
        raise ops.pebble.ChangeError("something", MagicMock())

    with ExitStack() as stack:
        # WHEN service restart fails
        stack.enter_context(patch("ops.model.Container.restart", new=raise_change_error))

        # THEN the charm errors out
        # technically an ops.pebble.ChangeError but the context manager doesn't catch it for some reason
        stack.enter_context(pytest.raises(Exception))
        ctx.run(ctx.on.pebble_ready(container), testing.State(containers={container}))


@pytest.mark.parametrize(
    "remote_databag, expected",
    (
        (
            {
                "remote_write_endpoints": json.dumps([{"url": "test-url.com"}]),
                "worker_config": json.dumps("test"),
            },
            [{"url": "test-url.com"}],
        ),
        ({"remote_write_endpoints": json.dumps(None), "worker_config": json.dumps("test")}, []),
        (
            {
                "remote_write_endpoints": json.dumps(
                    [{"url": "test-url.com"}, {"url": "test2-url.com"}]
                ),
                "worker_config": json.dumps("test"),
            },
            [{"url": "test-url.com"}, {"url": "test2-url.com"}],
        ),
    ),
)
def test_get_remote_write_endpoints(remote_databag, expected):
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    container = testing.Container(
        "foo",
        execs={testing.Exec(("update-ca-certificates", "--fresh"))},
        can_connect=True,
    )
    relation = testing.Relation(
        "cluster",
        remote_app_data=remote_databag,
    )
    with ctx(
        ctx.on.relation_changed(relation),
        testing.State(containers={container}, relations={relation}),
    ) as mgr:
        charm = mgr.charm
        mgr.run()
        assert charm.worker.cluster.get_remote_write_endpoints() == expected


@patch.object(Worker, "is_ready", new=lambda _: True)
@pytest.mark.parametrize(
    "ports",
    ([10, 11, 42], [2, 1232]),
)
def test_worker_ports(ports):
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    relation = testing.Relation(
        "cluster",
        remote_app_data={
            "worker_config": json.dumps(yaml.safe_dump({"testing": "config"})),
            "worker_ports": json.dumps(ports),
        },
    )
    with patch("ops.model.Unit.set_ports") as set_ports_patch:
        ctx.run(
            ctx.on.relation_changed(relation),
            testing.State(
                containers={testing.Container("foo", can_connect=True)}, relations={relation}
            ),
        )
        set_ports_patch.assert_called_with(*ports)


def test_config_preprocessor():
    # GIVEN a charm with a config preprocessor
    new_config = {"modified": "config"}

    class MyWorker(Worker):
        @property
        def _worker_config(self):
            # mock config processor that entirely replaces the config with another,
            # normally one would call super and manipulate
            return new_config

    class MyCharm(ops.CharmBase):
        layer = ops.pebble.Layer({"services": {"foo": {"command": ["bar"]}}})

        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            self.worker = MyWorker(
                self,
                "foo",
                lambda _: self.layer,
                {"cluster": "cluster"},
            )

    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={
            "options": {
                "role-all": {"type": "boolean", "default": "true"},
                "role-none": {"type": "boolean", "default": "false"},
            }
        },
    )

    # WHEN the charm writes the config to disk
    state_out = ctx.run(
        ctx.on.config_changed(),
        testing.State(
            config={"role-all": True},
            containers={
                testing.Container(
                    "foo",
                    can_connect=True,
                    execs={testing.Exec(("update-ca-certificates", "--fresh"))},
                )
            },
            relations={
                testing.Relation(
                    "cluster",
                    remote_app_data={
                        "worker_config": json.dumps(yaml.safe_dump({"original": "config"}))
                    },
                )
            },
        ),
    )

    # THEN the data gets preprocessed
    fs = Path(str(state_out.get_container("foo").get_filesystem(ctx)) + CONFIG_FILE)
    assert fs.read_text() == yaml.safe_dump(new_config)


@patch.object(Worker, "_update_worker_config", MagicMock(return_value=False))
@patch.object(Worker, "_set_pebble_layer", MagicMock(return_value=False))
@patch.object(Worker, "restart")
def test_worker_does_not_restart(restart_mock, tmp_path):
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    relation = testing.Relation(
        "cluster",
        remote_app_data={
            "worker_config": json.dumps("some: yaml"),
        },
    )
    # WHEN the charm receives any event and there are no changes to the config or the layer,
    #  but some of the services are down
    container = testing.Container(
        "foo",
        can_connect=True,
    )
    ctx.run(ctx.on.update_status(), testing.State(containers={container}, relations={relation}))

    assert not restart_mock.called


@patch.object(Worker, "_update_worker_config", MagicMock(return_value=False))
@patch.object(Worker, "_set_pebble_layer", MagicMock(return_value=False))
@patch.object(Worker, "restart")
def test_worker_does_not_restart_on_no_cert_changed(restart_mock, tmp_path):
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    secret = testing.Secret(
        {"private-key": "private"},
        label="private_id",
        owner="app",
    )
    relation = testing.Relation(
        "cluster",
        remote_app_data={
            "worker_config": json.dumps("some: yaml"),
            "ca_cert": json.dumps("ca"),
            "server_cert": json.dumps("cert"),
            "privkey_secret_id": json.dumps(secret.id),
            "s3_tls_ca_chain": json.dumps("s3_ca"),
        },
    )

    cert = tmp_path / "cert.cert"
    key = tmp_path / "key.key"
    client_ca = tmp_path / "client_ca.cert"
    s3_ca_chain = tmp_path / "s3_ca_chain.cert"
    root_ca_mocked_path = tmp_path / "rootcacert"

    cert.write_text("cert")
    key.write_text("private")
    client_ca.write_text("ca")
    s3_ca_chain.write_text("s3_ca")

    container = testing.Container(
        "foo",
        can_connect=True,
        execs={testing.Exec(("update-ca-certificates", "--fresh"))},
        mounts={
            "cert": testing.Mount(location=CERT_FILE, source=cert),
            "key": testing.Mount(location=KEY_FILE, source=key),
            "client_ca": testing.Mount(location=CLIENT_CA_FILE, source=client_ca),
            "s3_ca_chain": testing.Mount(location=S3_TLS_CA_CHAIN_FILE, source=s3_ca_chain),
            "root_ca": testing.Mount(location=root_ca_mocked_path, source=client_ca),
        },
    )

    ctx.run(
        ctx.on.update_status(),
        testing.State(leader=True, containers={container}, relations={relation}, secrets={secret}),
    )

    assert restart_mock.call_count == 0


@k8s_patch(is_ready=False)
@patch.object(Worker, "_update_config")
def test_worker_no_reconcile_when_patch_not_ready(_update_config_mock):
    class MyCharmWithResources(ops.CharmBase):
        layer = ops.pebble.Layer("")

        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            self.worker = Worker(
                self,
                "foo",
                lambda _: self.layer,
                {"cluster": "cluster"},
                readiness_check_endpoint="http://localhost:3200/ready",
                resources_requests=lambda _: {"cpu": "50m", "memory": "50Mi"},
                container_name="charm",
            )

    ctx = testing.Context(
        MyCharmWithResources,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )

    ctx.run(
        ctx.on.update_status(),
        testing.State(leader=True, containers={testing.Container("foo")}),
    )

    assert not _update_config_mock.called


@patch.object(Worker, "_update_worker_config", MagicMock(return_value=False))
@patch.object(Worker, "_set_pebble_layer", MagicMock(return_value=False))
@patch.object(Worker, "restart")
def test_worker_certs_update(restart_mock, tmp_path):
    # GIVEN a worker with no cert files on disk, and a cluster relation giving us some cert data
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    secret = testing.Secret(
        {"private-key": "private"},
        label="private_id",
        owner="app",
    )
    relation = testing.Relation(
        "cluster",
        remote_app_data={
            "worker_config": json.dumps("some: yaml"),
            "ca_cert": json.dumps("ca"),
            "server_cert": json.dumps("cert"),
            "privkey_secret_id": json.dumps(secret.id),
            "s3_tls_ca_chain": json.dumps("s3_ca"),
        },
    )

    cert = tmp_path / "cert.cert"
    key = tmp_path / "key.key"
    client_ca = tmp_path / "client_ca.cert"
    s3_ca_chain = tmp_path / "s3_ca_chain.cert"

    container = testing.Container(
        "foo",
        can_connect=True,
        execs={testing.Exec(("update-ca-certificates", "--fresh"))},
        mounts={
            "cert": testing.Mount(location=CERT_FILE, source=cert),
            "key": testing.Mount(location=KEY_FILE, source=key),
            "client_ca": testing.Mount(location=CLIENT_CA_FILE, source=client_ca),
            "s3_ca_chain": testing.Mount(location=S3_TLS_CA_CHAIN_FILE, source=s3_ca_chain),
        },
    )

    # WHEN the charm receives any event
    ctx.run(
        ctx.on.update_status(),
        testing.State(leader=True, containers={container}, relations={relation}, secrets={secret}),
    )

    # THEN the worker writes all tls data to the right locations on the container filesystem
    assert cert.read_text() == "cert"
    assert key.read_text() == "private"
    assert client_ca.read_text() == "ca"
    assert s3_ca_chain.read_text() == "s3_ca"

    # AND the worker restarts the workload
    assert restart_mock.call_count == 1


@patch.object(Worker, "_update_worker_config", MagicMock(return_value=False))
@patch.object(Worker, "_set_pebble_layer", MagicMock(return_value=False))
@patch.object(Worker, "restart")
@pytest.mark.parametrize("s3_ca_on_disk", (True, False))
def test_worker_certs_update_only_s3(restart_mock, tmp_path, s3_ca_on_disk):
    # GIVEN a worker with a tls-encrypted s3 bucket
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    relation = testing.Relation(
        "cluster",
        remote_app_data={
            "worker_config": json.dumps("some: yaml"),
            "s3_tls_ca_chain": json.dumps("s3_ca"),
        },
    )

    cert = tmp_path / "cert.cert"
    key = tmp_path / "key.key"
    client_ca = tmp_path / "client_ca.cert"
    s3_ca_chain = tmp_path / "s3_ca_chain.cert"
    if s3_ca_on_disk:
        s3_ca_chain.write_text("s3_ca")

    container = testing.Container(
        "foo",
        can_connect=True,
        execs={testing.Exec(("update-ca-certificates", "--fresh"))},
        mounts={
            "cert": testing.Mount(location=CERT_FILE, source=cert),
            "key": testing.Mount(location=KEY_FILE, source=key),
            "client_ca": testing.Mount(location=CLIENT_CA_FILE, source=client_ca),
            "s3_ca_chain": testing.Mount(location=S3_TLS_CA_CHAIN_FILE, source=s3_ca_chain),
        },
    )

    # WHEN the charm receives any event
    ctx.run(
        ctx.on.update_status(),
        testing.State(leader=True, containers={container}, relations={relation}),
    )

    # THEN the worker writes all tls data to the right locations on the container filesystem
    assert not cert.exists()
    assert not key.exists()
    assert not client_ca.exists()
    assert s3_ca_chain.read_text() == "s3_ca"

    # AND the worker restarts the workload IF it was not on disk already
    assert restart_mock.call_count == (0 if s3_ca_on_disk else 1)


@patch.object(Worker, "restart")
@patch.object(Worker, "stop")
@pytest.mark.parametrize("tls", (True, False))
def test_stop_called_on_no_cluster(stop_mock, restart_mock, tmp_path, tls):
    # GIVEN a worker who's all happy to begin with
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    cert = tmp_path / "cert.cert"
    key = tmp_path / "key.key"
    client_ca = tmp_path / "client_ca.cert"
    s3_ca_chain = tmp_path / "s3_ca_chain.cert"

    if tls:
        s3_ca_chain.write_text("something_tls")
        cert.write_text("something_tls")
        key.write_text("something_tls")
        client_ca.write_text("something_tls")

    container = testing.Container(
        "foo",
        can_connect=True,
        execs={testing.Exec(("update-ca-certificates", "--fresh"))},
        mounts={
            "cert": testing.Mount(location=CERT_FILE, source=cert),
            "key": testing.Mount(location=KEY_FILE, source=key),
            "client_ca": testing.Mount(location=CLIENT_CA_FILE, source=client_ca),
            "s3_ca_chain": testing.Mount(location=S3_TLS_CA_CHAIN_FILE, source=s3_ca_chain),
        },
    )

    # WHEN the charm receives any event
    ctx.run(
        ctx.on.update_status(),
        testing.State(leader=True, containers={container}),
    )

    fs = container.get_filesystem(ctx)
    # THEN the worker wipes all certificates if they are there
    assert not fs.joinpath(CERT_FILE).exists()
    assert not fs.joinpath(KEY_FILE).exists()
    assert not fs.joinpath(CLIENT_CA_FILE).exists()
    assert not fs.joinpath(S3_TLS_CA_CHAIN_FILE).exists()

    # AND the worker stops the workload instead of restarting it
    assert not restart_mock.called
    assert stop_mock.called


@patch.object(Worker, "is_ready", new=lambda _: False)
def test_worker_stop_all_services_if_not_ready(tmp_path):
    # GIVEN a worker with some services
    MyCharm.layer = ops.pebble.Layer(
        {
            "services": {
                "foo": {
                    "summary": "foos all the things",
                    "description": "bar",
                    "startup": "enabled",
                    "override": "merge",
                    "command": "ls -la",
                },
                "bar": {
                    "summary": "bars the foos",
                    "description": "bar",
                    "startup": "enabled",
                    "command": "exit 1",
                },
                "baz": {
                    "summary": "bazzes all of the bars",
                    "description": "bar",
                    "startup": "enabled",
                    "command": "echo hi",
                },
            }
        }
    )
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    # WHEN the charm receives any event, but it is not ready
    cfg = tmp_path / "cfg.yaml"
    cfg.write_text("some: yaml")
    container = testing.Container(
        "foo",
        layers={"base": MyCharm.layer},
        can_connect=True,
        mounts={"local": testing.Mount(location=CONFIG_FILE, source=cfg)},
        execs={
            testing.Exec(("update-ca-certificates", "--fresh")),
            testing.Exec(("/bin/foo", "-version"), stdout="foo"),
        },
        service_statuses={
            "foo": ops.pebble.ServiceStatus.ACTIVE,
            "bar": ops.pebble.ServiceStatus.ACTIVE,
            "baz": ops.pebble.ServiceStatus.INACTIVE,
        },
    )
    state_out = ctx.run(ctx.on.pebble_ready(container), testing.State(containers={container}))

    # THEN the charm restarts all the services that are down
    container_out = state_out.get_container("foo")
    service_statuses = container_out.service_statuses.values()
    assert all(svc is ops.pebble.ServiceStatus.INACTIVE for svc in service_statuses), [
        stat.value for stat in service_statuses
    ]


@patch("socket.getfqdn")
def test_invalid_url(mock_socket_fqdn):
    # Test that when socket returns an invalid url as a Fully Qualified Domain Name,
    #   ClusterRequirer.publish_unit_address raises a ValueError exception

    # GIVEN a properly configured charm
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )

    # AND ClusterRequirer is passed an invalid url as FQDN
    mock_socket_fqdn.return_value = "http://www.invalid-]url.com"

    # WHEN the charm executes any event
    # THEN the charm raises an error with the appropriate cause
    with pytest.raises(UncaughtCharmError) as exc:
        ctx.run(ctx.on.update_status(), testing.State(containers={testing.Container("foo")}))
    assert isinstance(exc.value.__cause__, ValueError)


@pytest.mark.parametrize(
    "remote_databag, expected",
    (
        (
            {
                "charm_tracing_receivers": json.dumps({"url": "test-url.com"}),
                "worker_config": json.dumps("test"),
            },
            {"url": "test-url.com"},
        ),
        (
            {"charm_tracing_receivers": json.dumps(None), "worker_config": json.dumps("test")},
            {},
        ),
    ),
)
def test_get_charm_tracing_receivers(remote_databag, expected):
    # Test that when a relation changes the correct charm_tracing_receivers
    #   are returned by the ClusterRequirer

    # GIVEN a charm with a relation
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    container = testing.Container(
        "foo",
        execs={testing.Exec(("update-ca-certificates", "--fresh"))},
        can_connect=True,
    )

    relation = testing.Relation(
        "cluster",
        remote_app_data=remote_databag,
    )

    # WHEN the relation changes
    with ctx(
        ctx.on.relation_changed(relation),
        testing.State(containers={container}, relations={relation}),
    ) as mgr:
        charm = mgr.charm
        # THEN the charm tracing receivers are picked up correctly
        assert charm.worker.cluster.get_charm_tracing_receivers() == expected


@pytest.mark.parametrize(
    "remote_databag, expected",
    (
        (
            {
                "workload_tracing_receivers": json.dumps({"url": "test-url.com"}),
                "worker_config": json.dumps("test"),
            },
            {"url": "test-url.com"},
        ),
        (
            {"workload_tracing_receivers": json.dumps(None), "worker_config": json.dumps("test")},
            {},
        ),
    ),
)
def test_get_workload_tracing_receivers(remote_databag, expected):
    # Test that when a relation changes the correct workload_tracing_receivers
    #   are returned by the ClusterRequirer

    # GIVEN a charm with a relation
    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    container = testing.Container(
        "foo",
        execs={testing.Exec(("update-ca-certificates", "--fresh"))},
        can_connect=True,
    )

    relation = testing.Relation(
        "cluster",
        remote_app_data=remote_databag,
    )

    # WHEN the relation changes
    with ctx(
        ctx.on.relation_changed(relation),
        testing.State(containers={container}, relations={relation}),
    ) as mgr:
        charm = mgr.charm
        # THEN the charm tracing receivers are picked up correctly
        assert charm.worker.cluster.get_workload_tracing_receivers() == expected


@pytest.mark.parametrize(
    "event",
    (
        CharmEvents.update_status(),
        CharmEvents.start(),
        CharmEvents.install(),
        CharmEvents.relation_changed,
    ),
)
def test_worker_blocks_on_tls_misconfigured(event):
    class _MyCharm(MyCharm):
        layer = ops.pebble.Layer({"services": {"foo": {"command": ["bar"]}}})

    ctx = testing.Context(
        _MyCharm,
        meta={
            "name": "foo",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"foo": {"type": "oci-image"}},
        },
        config={"options": {"role-all": {"type": "boolean", "default": True}}},
    )
    container = testing.Container(
        "foo",
        execs={testing.Exec(("update-ca-certificates", "--fresh"))},
        can_connect=True,
    )

    # GIVEN the coordinator is giving us AT LEAST ONE https endpoint for charm_tracing
    #  OR workload tracing, but no cert
    relation = testing.Relation(
        "cluster",
        remote_app_data={
            "charm_tracing_receivers": json.dumps(
                {
                    "jaeger_thrift_http": "https://198.18.0.0:14268",
                    "otlp_http": "http://198.18.0.0:4318",
                }
            ),
            "workload_tracing_receivers": json.dumps(
                {
                    "jaeger_thrift_http": "http://198.18.0.0:14268",
                    "otlp_http": "http://198.18.0.0:4318",
                }
            ),
            "worker_config": json.dumps("test"),
        },
    )

    if callable(event):
        event = event(relation)

    # WHEN any event occurs
    state_out = ctx.run(
        event,
        testing.State(containers={container}, relations={relation}),
    )
    # THEN the charm sets blocked
    assert isinstance(state_out.unit_status, BlockedStatus)
    assert "TLS misconfigured" in state_out.unit_status.message



import json

import pytest
from ops import CharmBase, Framework
from scenario import Context, Relation, State
from scenario.errors import UncaughtCharmError

from cosl.interfaces.datasource_exchange import (
    DatasourceExchange,
    DSExchangeAppData,
    EndpointValidationError,
    GrafanaDatasource,
)


@pytest.mark.parametrize(
    "meta, declared",
    (
        (
            {},
            (None, None),
        ),
        (
            {
                "requires": {"boo": {"interface": "gibberish"}},
                "provides": {"far": {"interface": "grafana_datasource_exchange"}},
            },
            ("far", "boo"),
        ),
        (
            {
                "requires": {"boo": {"interface": "grafana_datasource_exchange"}},
                "provides": {"goo": {"interface": "grafana_datasource_exchange"}},
            },
            ("far", "boo"),
        ),
    ),
)
def test_endpoint_validation_failure(meta, declared):
    # GIVEN a charm with this metadata and declared provider/requirer endpoints

    class BadCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            prov, req = declared
            self.ds_exchange = DatasourceExchange(
                self, provider_endpoint=prov, requirer_endpoint=req
            )

    # WHEN any event is processed
    with pytest.raises(UncaughtCharmError) as e:
        ctx = Context(BadCharm, meta={"name": "bob", **meta})
        ctx.run(ctx.on.update_status(), State())

    # THEN we raise an EndpointValidationError
    assert isinstance(e.value.__cause__, EndpointValidationError)


@pytest.mark.parametrize(
    "meta, declared",
    (
        (
            {
                "requires": {"boo": {"interface": "grafana_datasource_exchange"}},
                "provides": {"far": {"interface": "grafana_datasource_exchange"}},
            },
            ("far", "boo"),
        ),
        (
            {
                "provides": {"far": {"interface": "grafana_datasource_exchange"}},
            },
            ("far", None),
        ),
        (
            {
                "requires": {"boo": {"interface": "grafana_datasource_exchange"}},
            },
            (None, "boo"),
        ),
    ),
)
def test_endpoint_validation_ok(meta, declared):
    # GIVEN a charm with this metadata and declared provider/requirer endpoints
    class BadCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            prov, req = declared
            self.ds_exchange = DatasourceExchange(
                self, provider_endpoint=prov, requirer_endpoint=req
            )

    # WHEN any event is processed
    ctx = Context(BadCharm, meta={"name": "bob", **meta})
    ctx.run(ctx.on.update_status(), State())
    # THEN no exception is raised


def test_ds_publish():
    # GIVEN a charm with a single datasource_exchange relation
    class MyCharm(CharmBase):
        META = {
            "name": "robbie",
            "provides": {"foo": {"interface": "grafana_datasource_exchange"}},
        }

        def __init__(self, framework: Framework):
            super().__init__(framework)
            self.ds_exchange = DatasourceExchange(
                self, provider_endpoint="foo", requirer_endpoint=None
            )
            self.ds_exchange.publish([{"type": "tempo", "uid": "123", "grafana_uid": "123123"}])

    ctx = Context(MyCharm, meta=MyCharm.META)

    dse_in = Relation("foo")
    state_in = State(relations={dse_in}, leader=True)

    # WHEN we receive any event
    state_out = ctx.run(ctx.on.update_status(), state_in)

    # THEN we publish in our app databags any datasources we're aware of
    dse_out = state_out.get_relation(dse_in.id)
    assert dse_out.local_app_data
    data = DSExchangeAppData.load(dse_out.local_app_data)
    assert data.datasources[0].type == "tempo"
    assert data.datasources[0].uid == "123"


def test_ds_receive():
    # GIVEN a charm with a single datasource_exchange relation
    class MyCharm(CharmBase):
        META = {
            "name": "robbie",
            "provides": {"foo": {"interface": "grafana_datasource_exchange"}},
            "requires": {"bar": {"interface": "grafana_datasource_exchange"}},
        }

        def __init__(self, framework: Framework):
            super().__init__(framework)
            self.ds_exchange = DatasourceExchange(
                self, provider_endpoint="foo", requirer_endpoint="bar"
            )

    ctx = Context(MyCharm, meta=MyCharm.META)

    ds_requirer_in = [
        {"type": "c", "uid": "3", "grafana_uid": "4"},
        {"type": "a", "uid": "1", "grafana_uid": "5"},
        {"type": "b", "uid": "2", "grafana_uid": "6"},
    ]
    ds_provider_in = [{"type": "d", "uid": "4", "grafana_uid": "7"}]

    dse_requirer_in = Relation(
        "foo",
        remote_app_data=DSExchangeAppData(
            datasources=json.dumps(sorted(ds_provider_in, key=lambda raw_ds: raw_ds["uid"]))
        ).dump(),
    )
    dse_provider_in = Relation(
        "bar",
        remote_app_data=DSExchangeAppData(
            datasources=json.dumps(sorted(ds_requirer_in, key=lambda raw_ds: raw_ds["uid"]))
        ).dump(),
    )
    state_in = State(relations={dse_requirer_in, dse_provider_in}, leader=True)

    # WHEN we receive any event
    with ctx(ctx.on.update_status(), state_in) as mgr:
        # THEN we can access all datasources we're given
        dss = mgr.charm.ds_exchange.received_datasources
        assert [ds.type for ds in dss] == list("abcd")
        assert [ds.uid for ds in dss] == list("1234")
        assert isinstance(dss[0], GrafanaDatasource)



# Copyright 2022 Canonical Ltd.
# See LICENSE file for licensing details.
import unittest
import uuid
from collections import OrderedDict
from typing import Dict, List

from cosl.juju_topology import InvalidUUIDError, JujuTopology


class TestJujuTopology(unittest.TestCase):
    def setUp(self):
        self.input = {
            "model": "some-model",
            "model_uuid": "00000000-0000-4000-8000-000000000000",
            "application": "test-application",
            "unit": "test-application/0",
            "charm_name": "test-application",
        }
        self.topology = JujuTopology(
            self.input["model"],
            self.input["model_uuid"],
            self.input["application"],
            self.input["unit"],
            self.input["charm_name"],
        )

    def test_dict_includes_all_expected_keys(self):
        self.assertEqual(self.topology.as_dict(), self.input)

    def test_dict_does_not_include_excluded(self):
        expected = self.input.copy()
        expected = _filter_dict(expected, ["unit"])
        self.assertEqual(self.topology.as_dict(excluded_keys=["unit"]), expected)

    def test_dict_handles_remapping(self):
        expected = OrderedDict(
            ("name", v) if k == "charm_name" else (k, v) for k, v in self.input.items()
        )
        self.assertEqual(
            self.topology.as_dict(remapped_keys={"charm_name": "name"}),
            expected,
        )

    def test_identifier_omits_unit(self):
        expected = self.input.copy()
        expected = _filter_dict(expected, ["unit", "charm_name"])
        expected["model_uuid"] = expected["model_uuid"][:8]

        self.assertEqual(
            self.topology.identifier,
            "_".join([v for _, v in expected.items() if v]),
        )

    def test_label_matchers_dict_prefixes_keys(self):
        expected = OrderedDict(
            ("juju_{}".format("charm" if key == "charm_name" else key), val)
            for key, val in self.input.items()
        )
        expected = _filter_dict(expected, ["juju_unit"])

        self.assertEqual(self.topology.label_matcher_dict, expected)

    def test_label_matchers_creates_a_valid_matcher(self):
        expected = 'juju_model="some-model", juju_model_uuid="00000000-0000-4000-8000-000000000000", juju_application="test-application", juju_charm="test-application"'
        self.assertEqual(expected, self.topology.label_matchers)

    def test_from_dict(self):
        topology = JujuTopology.from_dict(self.input)
        self.assertEqual(topology.as_dict(), self.input)

    def test_invalid_uuid(self):
        """Test invalid UUIDs that doesn't match the regex we use."""
        t_uuid = str(uuid.uuid4()).split("-")
        block1 = t_uuid[0]
        block2 = t_uuid[1]
        block3 = t_uuid[2]
        block4 = t_uuid[3]
        block5 = t_uuid[4]
        sep = "-"

        invalid_uuids = [
            # Invalid character in each block
            sep.join([block1[:-1] + "ñ", block2, block3, block4, block5]),
            sep.join([block1, block2[:-1] + "ñ", block3, block4, block5]),
            sep.join([block1, block2, block3[:-1] + "ñ", block4, block5]),
            sep.join([block1, block2, block3, block4[:-1] + "ñ", block5]),
            sep.join([block1, block2, block3, block4, block5[:-1] + "ñ"]),
            # More characters in each block
            sep.join([block1 + "a", block2, block3, block4, block5]),
            sep.join([block1, block2 + "a", block3, block4, block5]),
            sep.join([block1, block2, block3 + "a", block4, block5]),
            sep.join([block1, block2, block3, block4 + "a", block5]),
            sep.join([block1, block2, block3, block4, block5 + "a"]),
            # Less characters in each block
            sep.join([block1[:-1], block2, block3, block4, block5]),
            sep.join([block1, block2[:-1], block3, block4, block5]),
            sep.join([block1, block2, block3[:-1], block4, block5]),
            sep.join([block1, block2, block3, block4[:-1], block5]),
            sep.join([block1, block2, block3, block4, block5[:-1]]),
            # More than one - in separator
            (sep * 2).join([block1, block2, block3, block4, block5]),
            # Missing blocks
            sep.join([block2, block3, block4, block5]),
            sep.join([block1, block3, block4, block5]),
            sep.join([block1, block2, block4, block5]),
            sep.join([block1, block2, block3, block5]),
            sep.join([block1, block2, block3, block4]),
            # UUID v4 validation.
            # 1st position in block3 must start with 4
            sep.join([block1, block2, "5" + block3[1:], block4, block5]),
            # 1st position in block4 must start with a, b, 8 or 9
            sep.join([block1, block2, block3, "w" + block4[1:], block5]),
            # UUID v1
            str(uuid.uuid1()),
            # UUID v5
            str(uuid.uuid5(uuid.NAMESPACE_DNS, "juju.is")),
        ]

        for invalid_uuid in invalid_uuids:
            topology_invalid_uuid = OrderedDict(
                [
                    ("model", "some-model"),
                    ("model_uuid", invalid_uuid),
                    ("application", "test-application"),
                    ("unit", "test-application/0"),
                    ("charm_name", "test-application"),
                ]
            )
            with self.assertRaises(InvalidUUIDError) as context:
                JujuTopology.from_dict(topology_invalid_uuid)

            self.assertEqual(f"'{invalid_uuid}' is not a valid UUID.", str(context.exception))


def _filter_dict(labels: Dict[str, str], excluded_keys: List[str]) -> "OrderedDict[str, str]":
    return OrderedDict({k: v for k, v in labels.items() if k not in excluded_keys})



#!/usr/bin/env python3
# Copyright 2024 Canonical
# See LICENSE file for licensing details.

"""Shared utilities for the coordinator -> worker "cluster" interface.

As this relation is cluster-internal and not intended for third-party charms to interact with
`-coordinator-k8s`, its only user will be the -worker-k8s charm. As such,
it does not live in a charm lib as most other relation endpoint wrappers do.
"""

import collections
import json
import logging
from typing import (
    Any,
    Callable,
    Counter,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
)
from urllib.parse import urlparse

import ops
import pydantic
import yaml
from ops import EventSource, Object, ObjectEvents, RelationCreatedEvent
from typing_extensions import TypedDict

import cosl
import cosl.interfaces.utils

log = logging.getLogger("_cluster")

DEFAULT_ENDPOINT_NAME = "-cluster"


# =============
# | Interface |
# =============


class RemoteWriteEndpoint(TypedDict):
    """Type of the remote write endpoints to be passed to the worker through cluster relation data."""

    url: str


class ConfigReceivedEvent(ops.EventBase):
    """Event emitted when the "-cluster" provider has shared a new  config."""

    config: Dict[str, Any]
    """The  config."""

    def __init__(self, handle: ops.framework.Handle, config: Dict[str, Any]):
        super().__init__(handle)
        self.config = config

    def snapshot(self) -> Dict[str, Any]:
        """Used by the framework to serialize the event to disk.

        Not meant to be called by charm code.
        """
        return {"config": json.dumps(self.config)}

    def restore(self, snapshot: Dict[str, Any]):
        """Used by the framework to deserialize the event from disk.

        Not meant to be called by charm code.
        """
        self.relation = json.loads(snapshot["config"])  # noqa


class ClusterError(Exception):
    """Base class for exceptions raised by this module."""


class DatabagAccessPermissionError(ClusterError):
    """Raised when a follower attempts to write leader settings."""


class _Topology(pydantic.BaseModel):
    """Juju topology information."""

    application: str
    unit: str
    charm_name: str


class ClusterRequirerAppData(cosl.interfaces.utils.DatabagModel):
    """App data that the worker sends to the coordinator."""

    role: str


class ClusterRequirerUnitData(cosl.interfaces.utils.DatabagModel):
    """Unit data the worker sends to the coordinator."""

    juju_topology: _Topology
    address: str


class ClusterProviderAppData(cosl.interfaces.utils.DatabagModel):
    """App data that the coordinator sends to the worker."""

    ### worker node configuration
    worker_config: str
    """The whole worker workload configuration, whatever it is. E.g. yaml-encoded things."""

    ### self-monitoring stuff
    loki_endpoints: Optional[Dict[str, str]] = None
    """Endpoints to which the workload (and the worker charm) can push logs to."""
    charm_tracing_receivers: Optional[Dict[str, str]] = None
    """Endpoints to which the the worker charm can push charm traces to."""
    workload_tracing_receivers: Optional[Dict[str, str]] = None
    """Endpoints to which the the worker can push workload traces to."""
    remote_write_endpoints: Optional[List[RemoteWriteEndpoint]] = None
    """Endpoints to which the workload (and the worker charm) can push metrics to."""
    worker_ports: Optional[List[int]] = None
    """Ports that the worker should open. If not provided, the worker will open all the legacy ones."""

    ### TLS stuff
    ca_cert: Optional[str] = None
    server_cert: Optional[str] = None
    privkey_secret_id: Optional[str] = None
    s3_tls_ca_chain: Optional[str] = None


class TLSData(NamedTuple):
    """Section of the cluster data that concerns TLS information."""

    ca_cert: Optional[str]
    server_cert: Optional[str]
    privkey_secret_id: Optional[str]
    s3_tls_ca_chain: Optional[str]


class ClusterChangedEvent(ops.EventBase):
    """Event emitted when any "-cluster" relation event fires."""


class ClusterRemovedEvent(ops.EventBase):
    """Event emitted when the relation with the "-cluster" provider has been severed.

    Or when the relation data has been wiped.
    """


class ClusterProviderEvents(ObjectEvents):
    """Events emitted by the ClusterProvider "-cluster" endpoint wrapper."""

    changed = EventSource(ClusterChangedEvent)


class ClusterRequirerEvents(ObjectEvents):
    """Events emitted by the ClusterRequirer "-cluster" endpoint wrapper."""

    config_received = EventSource(ConfigReceivedEvent)
    created = EventSource(RelationCreatedEvent)
    removed = EventSource(ClusterRemovedEvent)


class ClusterProvider(Object):
    """``-cluster`` provider endpoint wrapper."""

    on = ClusterProviderEvents()  # type: ignore

    def __init__(
        self,
        charm: ops.CharmBase,
        roles: FrozenSet[str],
        meta_roles: Optional[Mapping[str, Iterable[str]]] = None,
        key: Optional[str] = None,
        endpoint: str = DEFAULT_ENDPOINT_NAME,
        worker_ports: Optional[Callable[[str], Sequence[int]]] = None,
    ):
        super().__init__(charm, key or endpoint)
        self._charm = charm
        self._roles = roles
        self._meta_roles = meta_roles or {}
        self._worker_ports = worker_ports
        self.juju_topology = cosl.JujuTopology.from_charm(self._charm)

        # filter out common unhappy relation states
        self._relations: List[ops.Relation] = [
            rel for rel in self.model.relations[endpoint] if (rel.app and rel.data)
        ]

        # we coalesce all -cluster-relation-* events into a single cluster-changed API.
        # the coordinator uses a common exit hook reconciler, that's why.
        self.framework.observe(self._charm.on[endpoint].relation_created, self._on_cluster_changed)
        self.framework.observe(self._charm.on[endpoint].relation_joined, self._on_cluster_changed)
        self.framework.observe(self._charm.on[endpoint].relation_changed, self._on_cluster_changed)
        self.framework.observe(
            self._charm.on[endpoint].relation_departed, self._on_cluster_changed
        )
        self.framework.observe(self._charm.on[endpoint].relation_broken, self._on_cluster_changed)

    def _on_cluster_changed(self, _: ops.EventBase) -> None:
        self.on.changed.emit()

    def grant_privkey(self, label: str) -> str:
        """Grant the secret containing the privkey to all relations, and return the secret ID."""
        secret = self.model.get_secret(label=label)
        for relation in self._relations:
            secret.grant(relation)
        # can't return secret.id because secret was obtained by label, and so
        # we don't have an ID unless we fetch it
        return secret.get_info().id

    def publish_data(
        self,
        worker_config: str,
        ca_cert: Optional[str] = None,
        server_cert: Optional[str] = None,
        s3_tls_ca_chain: Optional[str] = None,
        privkey_secret_id: Optional[str] = None,
        loki_endpoints: Optional[Dict[str, str]] = None,
        charm_tracing_receivers: Optional[Dict[str, str]] = None,
        workload_tracing_receivers: Optional[Dict[str, str]] = None,
        remote_write_endpoints: Optional[List[RemoteWriteEndpoint]] = None,
    ) -> None:
        """Publish the config to all related worker clusters."""
        for relation in self._relations:
            if relation and self._remote_data_ready(relation):
                # obtain the worker ports for this relation, given the role advertised by the remote
                if worker_ports := self._worker_ports:
                    _worker_ports = list(
                        worker_ports(ClusterRequirerAppData.load(relation.data[relation.app]).role)
                    )
                else:
                    _worker_ports = None

                local_app_databag = ClusterProviderAppData(
                    worker_config=worker_config,
                    loki_endpoints=loki_endpoints,
                    ca_cert=ca_cert,
                    server_cert=server_cert,
                    privkey_secret_id=privkey_secret_id,
                    charm_tracing_receivers=charm_tracing_receivers,
                    workload_tracing_receivers=workload_tracing_receivers,
                    remote_write_endpoints=remote_write_endpoints,
                    s3_tls_ca_chain=s3_tls_ca_chain,
                    worker_ports=_worker_ports,
                )
                local_app_databag.dump(relation.data[self.model.app])

    @property
    def has_workers(self) -> bool:
        """Return True if the coordinator is connected to any worker."""
        # we use the presence of relations instead of addresses, because we want this
        # check to fail early
        return bool(self._relations)

    def _expand_roles(self, role_string: str) -> Set[str]:
        """Expand the meta-roles from a comma-separated list of roles."""
        expanded_roles: Set[str] = set()
        for role in role_string.split(","):
            if role in self._meta_roles:
                expanded_roles.update(self._meta_roles[role])
            else:
                expanded_roles.update({role})
        return expanded_roles

    def gather_addresses_by_role(self) -> Dict[str, Set[str]]:
        """Go through the worker's unit databags to collect all the addresses published by the units, by role."""
        data: Dict[str, Set[str]] = collections.defaultdict(set)
        for relation in self._relations:

            if not relation.app:
                log.debug(f"skipped {relation} as .app is None")
                continue

            try:
                worker_app_data = ClusterRequirerAppData.load(relation.data[relation.app])
            except cosl.interfaces.utils.DataValidationError as e:
                log.info(f"invalid databag contents: {e}")
                continue

            for worker_unit in relation.units:
                try:
                    worker_data = ClusterRequirerUnitData.load(relation.data[worker_unit])
                    unit_address = worker_data.address
                    for role in self._expand_roles(worker_app_data.role):
                        data[role].add(unit_address)
                except cosl.interfaces.utils.DataValidationError as e:
                    log.info(f"invalid databag contents: {e}")
                    continue
        return data

    def gather_addresses(self) -> Tuple[str, ...]:
        """Go through the worker's unit databags to collect all the addresses published by the units."""
        data: Set[str] = set()
        addresses_by_role = self.gather_addresses_by_role()
        for _, address_set in addresses_by_role.items():
            data.update(address_set)

        return tuple(sorted(data))

    def gather_roles(self) -> Dict[str, int]:
        """Go through the worker's app databags and sum the available application roles."""
        data: Counter[str] = collections.Counter()
        for relation in self._relations:
            if relation.app:
                remote_app_databag = relation.data[relation.app]
                try:
                    worker_role: str = ClusterRequirerAppData.load(remote_app_databag).role
                except cosl.interfaces.utils.DataValidationError as e:
                    log.error(f"invalid databag contents: {e}")
                    continue

                # the number of units with each role is the number of remote units
                role_n = len(relation.units)  # exclude this unit
                for role in self._expand_roles(worker_role):
                    data[role] += role_n

        dct = dict(data)
        return dct

    def gather_topology(self) -> List[Dict[str, str]]:
        """Gather Juju topology by unit."""
        data: List[Dict[str, str]] = []
        for relation in self._relations:
            if not relation.app:
                continue

            for worker_unit in relation.units:
                try:
                    worker_data = ClusterRequirerUnitData.load(relation.data[worker_unit])
                    unit_address = worker_data.address
                except cosl.interfaces.utils.DataValidationError as e:
                    log.info(f"invalid databag contents: {e}")
                    continue
                worker_topology = {
                    "address": unit_address,
                    "application": worker_data.juju_topology.application,
                    "unit": worker_data.juju_topology.unit,
                    "charm_name": worker_data.juju_topology.charm_name,
                }
                data.append(worker_topology)

        return data

    def get_address_from_role(self, role: str) -> Optional[str]:
        """Get datasource address."""
        addresses_by_role = self.gather_addresses_by_role()
        if address_set := addresses_by_role.get(role, None):
            return address_set.pop()
        return None

    def _remote_data_ready(self, relation: ops.Relation) -> bool:
        """Verify that each worker unit and the worker leader have published their data to the cluster relation.

        - unit address is published
        - roles are published
        """
        if not relation.app or not relation.units or not relation.data:
            return False

        # check if unit data is published
        for worker_unit in relation.units:
            try:
                ClusterRequirerUnitData.load(relation.data[worker_unit])
            except cosl.interfaces.utils.DataValidationError:
                return False

        # check if app data is published
        if self._charm.unit.is_leader():
            try:
                ClusterRequirerAppData.load(relation.data[relation.app])
            except cosl.interfaces.utils.DataValidationError:
                return False

        return True


class ClusterRequirer(Object):
    """``-cluster`` requirer endpoint wrapper."""

    on = ClusterRequirerEvents()  # type: ignore

    def __init__(
        self,
        charm: ops.CharmBase,
        key: Optional[str] = None,
        endpoint: str = DEFAULT_ENDPOINT_NAME,
    ):
        super().__init__(charm, key or endpoint)
        self._charm = charm
        self.juju_topology = cosl.JujuTopology.from_charm(self._charm)

        relation = self.model.get_relation(endpoint)
        self.relation: Optional[ops.Relation] = (
            relation if relation and relation.app and relation.data else None
        )

        self.framework.observe(
            self._charm.on[endpoint].relation_changed, self._on_cluster_relation_changed  # type: ignore
        )
        self.framework.observe(
            self._charm.on[endpoint].relation_created, self._on_cluster_relation_created  # type: ignore
        )
        self.framework.observe(
            self._charm.on[endpoint].relation_broken, self._on_cluster_relation_broken  # type: ignore
        )

    def _on_cluster_relation_broken(self, _event: ops.RelationBrokenEvent):
        self.on.removed.emit()

    def _on_cluster_relation_created(self, event: ops.RelationCreatedEvent):
        self.on.created.emit(relation=event.relation, app=event.app, unit=event.unit)

    def _on_cluster_relation_changed(self, _event: ops.RelationChangedEvent):
        # to prevent the event from firing if the relation is in an unhealthy state (breaking...)
        if self.relation:
            new_config = self.get_worker_config()
            if new_config:
                self.on.config_received.emit(new_config)

            # if we have published our data, but we receive an empty/invalid config,
            # then the remote end must have removed it.
            elif self.is_published():
                self.on.removed.emit()

    def is_published(self):
        """Verify that the local side has done all they need to do.

        - unit address is published
        - roles are published
        """
        relation = self.relation
        if not relation:
            return False

        unit_data = relation.data[self._charm.unit]
        app_data = relation.data[self._charm.app]

        try:
            ClusterRequirerUnitData.load(unit_data)
            if self._charm.unit.is_leader():
                ClusterRequirerAppData.load(app_data)
        except cosl.interfaces.utils.DataValidationError as e:
            log.info(f"invalid databag contents: {e}")
            return False
        return True

    def publish_unit_address(self, url: str):
        """Publish this unit's URL via the unit databag."""
        try:
            urlparse(url)
        except Exception as e:
            raise ValueError(f"{url} is an invalid url") from e

        databag_model = ClusterRequirerUnitData(
            juju_topology=dict(self.juju_topology.as_dict()),  # type: ignore
            address=url,
        )
        relation = self.relation
        if relation:
            unit_databag = relation.data[self.model.unit]  # type: ignore # all checks are done in __init__
            databag_model.dump(unit_databag)

    def publish_app_roles(self, roles: Iterable[str]):
        """Publish this application's roles via the application databag."""
        if not self._charm.unit.is_leader():
            raise DatabagAccessPermissionError("only the leader unit can publish roles.")

        relation = self.relation
        if relation:
            databag_model = ClusterRequirerAppData(role=",".join(roles))
            databag_model.dump(relation.data[self.model.app])

    def _get_data_from_coordinator(self) -> Optional[ClusterProviderAppData]:
        """Fetch the contents of the doordinator databag."""
        data: Optional[ClusterProviderAppData] = None
        relation = self.relation
        # TODO: does this need a leader guard ??? maybe?
        if relation:
            try:
                databag = relation.data[relation.app]  # type: ignore # all checks are done in __init__
                coordinator_databag = ClusterProviderAppData.load(databag)
                data = coordinator_databag
            except cosl.interfaces.utils.DataValidationError as e:
                log.info(f"invalid databag contents: {e}")
                return None  # explicit is better than implicit

        return data

    def get_worker_config(self) -> Dict[str, Any]:
        """Fetch the worker config from the coordinator databag."""
        data = self._get_data_from_coordinator()
        if data:
            return yaml.safe_load(data.worker_config)
        return {}

    def get_worker_ports(self) -> Optional[Tuple[int, ...]]:
        """Obtain, from the cluster relation, the ports that the worker should be opening."""
        data = self._get_data_from_coordinator()
        if data and data.worker_ports:
            return tuple(data.worker_ports)
        return

    def get_loki_endpoints(self) -> Dict[str, str]:
        """Fetch the loki endpoints from the coordinator databag."""
        data = self._get_data_from_coordinator()
        if data:
            return data.loki_endpoints or {}
        return {}

    def get_tls_data(self, allow_none: bool = False) -> Optional[TLSData]:
        """Fetch certificates and the private key secrets id for the worker config."""
        data = self._get_data_from_coordinator()
        if not data:
            return None

        if (
            not data.ca_cert or not data.server_cert or not data.privkey_secret_id
        ) and not allow_none:
            return None

        return TLSData(
            ca_cert=data.ca_cert,
            server_cert=data.server_cert,
            privkey_secret_id=data.privkey_secret_id,
            s3_tls_ca_chain=data.s3_tls_ca_chain,
        )

    def get_charm_tracing_receivers(self) -> Dict[str, str]:
        """Fetch the charm tracing receivers from the coordinator databag."""
        data = self._get_data_from_coordinator()
        if data:
            return data.charm_tracing_receivers or {}
        return {}

    def get_workload_tracing_receivers(self) -> Dict[str, str]:
        """Fetch the workload tracing receivers from the coordinator databag."""
        data = self._get_data_from_coordinator()
        if data:
            return data.workload_tracing_receivers or {}
        return {}

    def get_remote_write_endpoints(self) -> List[RemoteWriteEndpoint]:
        """Fetch the remote write endpoints from the coordinator databag."""
        data = self._get_data_from_coordinator()
        if data:
            return data.remote_write_endpoints or []
        return []



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.
# Original LokiEmitter/LokiHandler implementation from
# https://github.com/GreyZmeem/python-logging-loki (MIT licensed)

"""Loki logger."""

import copy
import functools
import json
import logging
import string
import time
import urllib.error
from logging.config import ConvertingDict
from typing import Any, Dict, Optional, Tuple, cast
from urllib import request

logger = logging.getLogger("loki-logger")

# prevent infinite recursion because on failure urllib3 will push more logs
# https://github.com/GreyZmeem/python-logging-loki/issues/18
logging.getLogger("urllib3").setLevel(logging.INFO)


class LokiEmitter:
    """Base Loki emitter class."""

    #: Success HTTP status code from Loki API.
    success_response_code: int = 204

    #: Label name indicating logging level.
    level_label: str = "severity"
    #: Label name indicating logger name.
    logger_label: str = "logger"

    #: String contains chars that can be used in label names in LogQL.
    label_allowed_chars: str = "".join((string.ascii_letters, string.digits, "_"))
    #: A list of pairs of characters to replace in the label name.
    label_replace_with: Tuple[Tuple[str, str], ...] = (
        ("'", ""),
        ('"', ""),
        (" ", "_"),
        (".", "_"),
        ("-", "_"),
    )

    def __init__(
        self, url: str, labels: Optional[Dict[str, str]] = None, cert: Optional[str] = None
    ):
        """Create new Loki emitter.

        Arguments:
            url: Endpoint used to send log entries to Loki (e.g.
            `https://my-loki-instance/loki/api/v1/push`).
            labels: Default labels added to every log record.
            cert: Absolute path to a ca cert for TLS authentication.

        """
        #: Tags that will be added to all records handled by this handler.
        self.labels: Dict[str, str] = labels or {}
        #: Loki JSON push endpoint (e.g `http://127.0.0.1/loki/api/v1/push`)
        self.url = url
        #: Optional cert for TLS auth
        self.cert = cert
        #: only notify once on push failure, to avoid spamming error logs
        self._error_notified_once = False

    def _send_request(self, req: request.Request, jsondata_encoded: bytes):
        return request.urlopen(req, jsondata_encoded, capath=self.cert)

    def __call__(self, record: logging.LogRecord, line: str):
        """Send log record to Loki."""
        payload = self.build_payload(record, line)
        req = request.Request(self.url, method="POST")
        req.add_header("Content-Type", "application/json; charset=utf-8")
        jsondata_encoded = json.dumps(payload).encode("utf-8")

        try:
            resp = self._send_request(req, jsondata_encoded)
        except urllib.error.HTTPError as e:
            if not self._error_notified_once:
                # set this BEFORE logging anything, or we'll recurse into a stack overflow!
                self._error_notified_once = True
                logger.error(f"error pushing logs to {self.url}: {e.code, e.reason}")  # type: ignore
            return

        if resp.getcode() != self.success_response_code:
            raise ValueError(
                "Unexpected Loki API response status code: {0}".format(resp.status_code)
            )

    def build_payload(self, record: logging.LogRecord, line: str) -> Dict[str, Any]:
        """Build JSON payload with a log entry."""
        labels = self.build_labels(record)
        ns = 1e9
        ts = str(int(time.time() * ns))
        stream = {
            "stream": labels,
            "values": [[ts, line]],
        }
        return {"streams": [stream]}

    @functools.lru_cache(256)
    def format_label(self, label: str) -> str:
        """Build label to match prometheus format.

        `Label format <https://prometheus.io/docs/concepts/data_model/#metric-names-and-labels>`_
        """
        for char_from, char_to in self.label_replace_with:
            label = label.replace(char_from, char_to)
        return "".join(char for char in label if char in self.label_allowed_chars)

    def build_labels(self, record: logging.LogRecord) -> Dict[str, str]:
        """Return labels that must be sent to Loki with a log record."""
        labels: Dict[str, str] = (
            dict(self.labels) if isinstance(self.labels, ConvertingDict) else self.labels
        )
        labels = cast(Dict[str, Any], copy.deepcopy(labels))
        labels[self.level_label] = record.levelname.lower()
        labels[self.logger_label] = record.name

        # if the user implemented a logrecord subclass with a .labels attributes, attempt to
        # respect it and add those labels on top of those registered on the LokiEmitter class.
        extra_labels: Any = getattr(record, "labels", {})
        if not isinstance(extra_labels, dict):
            return labels

        label_name: Any
        label_value: Any
        for label_name, label_value in extra_labels.items():
            if not isinstance(label_name, str) or not isinstance(label_value, str):
                return labels

            cleared_name = self.format_label(label_name)
            if cleared_name:
                labels[cleared_name] = label_value

        return labels


class LokiHandler(logging.Handler):
    """Log handler that sends log records to Loki.

    `Loki API <https://github.com/grafana/loki/blob/master/docs/api.md>`  # wokeignore:rule=master
    """

    def __init__(
        self,
        url: str,
        labels: Optional[Dict[str, str]] = None,
        # username, password tuple
        cert: Optional[str] = None,
    ):
        """Create new Loki logging handler.

        Arguments:
            url: Endpoint used to send log entries to Loki (e.g.
            `https://my-loki-instance/loki/api/v1/push`).
            labels: Default labels added to every log record.
            cert: Optional absolute path to cert file for TLS auth.

        """
        super().__init__()
        self.emitter = LokiEmitter(url, labels, cert)

    def emit(self, record: logging.LogRecord):
        """Send log record to Loki."""
        # noinspection PyBroadException
        try:
            self.emitter(record, self.format(record))
        except Exception:
            self.handleError(record)



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

import json
import unittest

from cosl import DashboardPath40UID, GrafanaDashboard, LZMABase64


class TestRoundTripEncDec(unittest.TestCase):
    """Tests the round-trip encoding/decoding of the GrafanaDashboard class."""

    def test_round_trip(self):
        d = {
            "some": "dict",
            "with": "keys",
            "even": [{"nested": "types", "and_integers": [42, 42]}],
        }
        self.assertDictEqual(d, GrafanaDashboard._serialize(json.dumps(d))._deserialize())


class TestLZMABase64(unittest.TestCase):
    """Tests the round-trip encoding/decoding of the GrafanaDashboard class."""

    def test_round_trip(self):
        s = "starting point"
        self.assertEqual(s, LZMABase64.decompress(LZMABase64.compress(s)))


class TestGenerateUID(unittest.TestCase):
    """Spec for the UID generation logic."""

    def test_uid_length_is_40(self):
        self.assertEqual(40, len(DashboardPath40UID.generate("my-charm", "my-dash.json")))

    def test_collisions(self):
        """A very naive and primitive collision check that is meant to catch trivial errors."""
        self.assertNotEqual(
            DashboardPath40UID.generate("some-charm", "dashboard1.json"),
            DashboardPath40UID.generate("some-charm", "dashboard2.json"),
        )

        self.assertNotEqual(
            DashboardPath40UID.generate("some-charm", "dashboard.json"),
            DashboardPath40UID.generate("diff-charm", "dashboard.json"),
        )

    def test_validity(self):
        """Make sure validity check fails for trivial cases."""
        self.assertFalse(DashboardPath40UID.is_valid("1234"))
        self.assertFalse(DashboardPath40UID.is_valid("short non-hex string"))
        self.assertFalse(DashboardPath40UID.is_valid("non-hex string, crafted to be 40 chars!!"))

        self.assertTrue(DashboardPath40UID.is_valid("0" * 40))
        self.assertTrue(
            DashboardPath40UID.is_valid(
                DashboardPath40UID.generate("some-charm", "dashboard.json")
            )
        )



# Copyright 2023 Canonical
# See LICENSE file for licensing details.

"""Workload manager for Nginx. Used by the coordinator to load-balance and group the workers."""

import logging
import subprocess
from pathlib import Path
from typing import Callable, Optional, TypedDict

from ops import CharmBase, pebble

logger = logging.getLogger(__name__)

# TODO: should we add these to _NginxMapping and make them configurable / accessible?
NGINX_DIR = "/etc/nginx"
NGINX_CONFIG = f"{NGINX_DIR}/nginx.conf"
KEY_PATH = f"{NGINX_DIR}/certs/server.key"
CERT_PATH = f"{NGINX_DIR}/certs/server.cert"
CA_CERT_PATH = "/usr/local/share/ca-certificates/ca.cert"

_NginxMapping = TypedDict(
    "_NginxMapping", {"nginx_port": int, "nginx_exporter_port": int}, total=True
)
NginxMappingOverrides = TypedDict(
    "NginxMappingOverrides", {"nginx_port": int, "nginx_exporter_port": int}, total=False
)
DEFAULT_OPTIONS: _NginxMapping = {
    "nginx_port": 8080,
    "nginx_exporter_port": 9113,
}


class Nginx:
    """Helper class to manage the nginx workload."""

    config_path = NGINX_CONFIG
    _name = "nginx"
    options: _NginxMapping = DEFAULT_OPTIONS

    def __init__(
        self,
        charm: CharmBase,
        config_getter: Callable[[], str],
        options: Optional[NginxMappingOverrides] = None,
    ):
        self._charm = charm
        self._config_getter = config_getter
        self._container = self._charm.unit.get_container("nginx")
        self.options.update(options or {})

    @property
    def are_certificates_on_disk(self) -> bool:
        """Return True if the certificates files are on disk."""
        return (
            self._container.can_connect()
            and self._container.exists(CERT_PATH)
            and self._container.exists(KEY_PATH)
            and self._container.exists(CA_CERT_PATH)
        )

    def configure_tls(self, private_key: str, server_cert: str, ca_cert: str) -> None:
        """Save the certificates file to disk and run update-ca-certificates."""
        if self._container.can_connect():
            # Read the current content of the files (if they exist)
            current_server_cert = (
                self._container.pull(CERT_PATH).read() if self._container.exists(CERT_PATH) else ""
            )
            current_private_key = (
                self._container.pull(KEY_PATH).read() if self._container.exists(KEY_PATH) else ""
            )
            current_ca_cert = (
                self._container.pull(CA_CERT_PATH).read()
                if self._container.exists(CA_CERT_PATH)
                else ""
            )

            if (
                current_server_cert == server_cert
                and current_private_key == private_key
                and current_ca_cert == ca_cert
            ):
                # No update needed
                return
            self._container.push(KEY_PATH, private_key, make_dirs=True)
            self._container.push(CERT_PATH, server_cert, make_dirs=True)
            self._container.push(CA_CERT_PATH, ca_cert, make_dirs=True)

            # push CA cert to charm container
            Path(CA_CERT_PATH).parent.mkdir(parents=True, exist_ok=True)
            Path(CA_CERT_PATH).write_text(ca_cert)

            # FIXME: uncomment as soon as the nginx image contains the ca-certificates package
            # self._container.exec(["update-ca-certificates", "--fresh"])

    def delete_certificates(self) -> None:
        """Delete the certificate files from disk and run update-ca-certificates."""
        if self._container.can_connect():
            if self._container.exists(CERT_PATH):
                self._container.remove_path(CERT_PATH, recursive=True)
            if self._container.exists(KEY_PATH):
                self._container.remove_path(KEY_PATH, recursive=True)
            if self._container.exists(CA_CERT_PATH):
                self._container.remove_path(CA_CERT_PATH, recursive=True)
            if Path(CA_CERT_PATH).exists():
                Path(CA_CERT_PATH).unlink(missing_ok=True)
            # FIXME: uncomment as soon as the nginx image contains the ca-certificates package
            # self._container.exec(["update-ca-certificates", "--fresh"])

    def _has_config_changed(self, new_config: str) -> bool:
        """Return True if the passed config differs from the one on disk."""
        if not self._container.can_connect():
            logger.debug("Could not connect to Nginx container")
            return False

        try:
            current_config = self._container.pull(self.config_path).read()
        except (pebble.ProtocolError, pebble.PathError) as e:
            logger.warning(
                "Could not check the current nginx configuration due to "
                "a failure in retrieving the file: %s",
                e,
            )
            return False

        return current_config != new_config

    def reload(self) -> None:
        """Reload the nginx config without restarting the service."""
        if self._container.can_connect():
            self._container.exec(["nginx", "-s", "reload"])

    def configure_pebble_layer(self) -> None:
        """Configure pebble layer."""
        if self._container.can_connect():
            new_config: str = self._config_getter()
            should_restart: bool = self._has_config_changed(new_config)
            self._container.push(self.config_path, new_config, make_dirs=True)  # type: ignore
            self._container.add_layer("nginx", self.layer, combine=True)
            self._container.autostart()

            if should_restart:
                logger.info("new nginx config: restarting the service")
                self.reload()

    @property
    def layer(self) -> pebble.Layer:
        """Return the Pebble layer for Nginx."""
        return pebble.Layer(
            {
                "summary": "nginx layer",
                "description": "pebble config layer for Nginx",
                "services": {
                    "nginx": {
                        "override": "replace",
                        "summary": "nginx",
                        "command": "nginx -g 'daemon off;'",
                        "startup": "enabled",
                    }
                },
            }
        )


class NginxPrometheusExporter:
    """Helper class to manage the nginx prometheus exporter workload."""

    options: _NginxMapping = DEFAULT_OPTIONS

    def __init__(self, charm: CharmBase, options: Optional[NginxMappingOverrides] = None) -> None:
        self._charm = charm
        self._container = self._charm.unit.get_container("nginx-prometheus-exporter")
        self.options.update(options or {})

    def configure_pebble_layer(self) -> None:
        """Configure pebble layer."""
        if self._container.can_connect():
            self._container.add_layer("nginx-prometheus-exporter", self.layer, combine=True)
            self._container.autostart()

    @property
    def are_certificates_on_disk(self) -> bool:
        """Return True if the certificates files are on disk."""
        return (
            self._container.can_connect()
            and self._container.exists(CERT_PATH)
            and self._container.exists(KEY_PATH)
            and self._container.exists(CA_CERT_PATH)
        )

    @property
    def layer(self) -> pebble.Layer:
        """Return the Pebble layer for Nginx Prometheus exporter."""
        scheme = "https" if self.are_certificates_on_disk else "http"  # type: ignore
        return pebble.Layer(
            {
                "summary": "nginx prometheus exporter layer",
                "description": "pebble config layer for Nginx Prometheus exporter",
                "services": {
                    "nginx": {
                        "override": "replace",
                        "summary": "nginx prometheus exporter",
                        "command": f"nginx-prometheus-exporter --no-nginx.ssl-verify --web.listen-address=:{self.options['nginx_exporter_port']}  --nginx.scrape-uri={scheme}://127.0.0.1:{self.options['nginx_port']}/status",
                        "startup": "enabled",
                    }
                },
            }
        )


def is_ipv6_enabled() -> bool:
    """Check if IPv6 is enabled on the container's network interfaces."""
    try:
        output = subprocess.run(
            ["ip", "-6", "address", "show"], check=True, capture_output=True, text=True
        )
    except subprocess.CalledProcessError:
        # if running the command failed for any reason, assume ipv6 is not enabled.
        return False
    return bool(output.stdout)
